/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.bridge;

import android.app.Application;
import android.content.ComponentCallbacks;
import android.content.Context;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageManager;
import android.content.pm.ProviderInfo;
import android.content.res.Configuration;
import android.content.res.Resources;
import android.os.Build;
import android.os.Process;
import android.support.multidex.MultiDex;
import android.taobao.atlas.framework.Atlas;
import android.taobao.atlas.framework.Framework;
import android.taobao.atlas.hack.AndroidHack;
import android.taobao.atlas.hack.AtlasHacks;
import android.taobao.atlas.runtime.AtlasPreLauncher;
import android.taobao.atlas.runtime.RuntimeVariables;

import android.taobao.atlas.startup.KernalVersionManager;
import android.taobao.atlas.util.SoLoader;
import android.taobao.atlas.util.log.IAlarmer;
import android.taobao.atlas.util.log.IMonitor;
import android.taobao.atlas.util.log.impl.AtlasAlarmer;
import android.taobao.atlas.util.log.impl.AtlasMonitor;
import android.text.TextUtils;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.WindowManager;

import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import dalvik.system.DexFile;

/**
 * Created by guanjie on 2017/1/26.
 */

public class BridgeApplicationDelegate {

    private String mRealApplicationName ;
    private Application mRealApplication;
    private Application mRawApplication;
    private String mInstalledVersionName;
    private String mCurrentProcessname;
    private long   mInstalledVersionCode;
    private long   mLastUpdateTime;
    private boolean mIsUpdated;
    private String mApkPath;
    private Object mdexLoadBooster;
    private List<ProviderInfo> mBoundApplication_provider;

    public BridgeApplicationDelegate(Application rawApplication,String processname,String installedVersion,
                                     long versioncode,long lastupdatetime,String apkPath,boolean isUpdated,Object dexLoadBooster){
//        if(Build.VERSION.SDK_INT<=19 && getClass().getClassLoader().getClass().getName().startsWith("com.ali.mobisecenhance")){
//            try {
//                Field pathListField = AndroidHack.findField(rawApplication.getClassLoader(), "pathList");
//                Object dexPathList = pathListField.get(rawApplication.getClassLoader());
//                Field elementsField = AndroidHack.findField(dexPathList,"dexElements");
//                Object[] elements = (Object[])elementsField.get(dexPathList);
//                Log.e("BridgeApplication","get Elements :"+ elements);
//
//                if(elements.length>0) {
//                    Field dexFileField = elements[0].getClass().getDeclaredField("dexFile");
//                    dexFileField.setAccessible(true);
//                    for(int x=elements.length-1; x>=0; x--){
//                        DexFile dexFile = (DexFile) dexFileField.get(elements[x]);
//                        if(dexFile.getName().contains("com.taobao.maindex")) {
//                            //针对动态部署处理过的dex做判断
//                            boolean findDexToDelete = false;
//                            Enumeration<String> enumeration = dexFile.entries();
//                            while (enumeration.hasMoreElements()) {
//                                if (enumeration.nextElement().replace("/", ".").startsWith("com.ali.mobisecenhance.ld.util")) {
//                                    findDexToDelete = true;
//                                    break;
//                                }
//                            }
//                            if(findDexToDelete){
//                                Log.e("BridgeApplication","delete dexfile :"+dexFile.getName());
//                                dexFileField.set(elements[x],null);
//                                break;
//                            }
//                        }
//                    }
//                }
//            }catch(Throwable e){
//                e.printStackTrace();
//            }
//        }

        mRawApplication = rawApplication;
        mCurrentProcessname = processname;
        mInstalledVersionName = installedVersion;
        mInstalledVersionCode = versioncode;
        mLastUpdateTime = lastupdatetime;
        mIsUpdated = isUpdated;
        mApkPath = apkPath;
        mdexLoadBooster = dexLoadBooster;
//        PackageManagerDelegate.delegatepackageManager(rawApplication.getBaseContext());
    }

    public void attachBaseContext(){
        AtlasHacks.defineAndVerify();
        RuntimeVariables.androidApplication = mRawApplication;
        RuntimeVariables.originalResources = mRawApplication.getResources();
        RuntimeVariables.sCurrentProcessName = mCurrentProcessname;
        RuntimeVariables.sInstalledVersionCode = mInstalledVersionCode;
        RuntimeVariables.sAppLastUpdateTime = mLastUpdateTime;
        RuntimeVariables.sApkPath = mApkPath;
        RuntimeVariables.delegateResources = mRawApplication.getResources();
        RuntimeVariables.sDexLoadBooster = mdexLoadBooster;
        Log.e("BridgeApplication","length =" + new File(mRawApplication.getApplicationInfo().sourceDir).length());

//        if(Build.MANUFACTURER.equalsIgnoreCase("vivo") && Build.VERSION.SDK_INT== 23) {
//            ;
////            try {
////                File appSGLib = mRawApplication.getDir("SGLib", Context.MODE_PRIVATE);
////                File mark = new File(mRawApplication.getFilesDir(), "vivo_appSGLib_mark");
////                if (appSGLib.exists() && !mark.exists()) {
////                    mark.createNewFile();
////                    File[] files = appSGLib.listFiles();
////                    for(File file : files){
////                        if(file.exists() && file.isDirectory() && file.getName().startsWith("app_")){
////                            deleteDirectory(file);
////                        }
////                    }
////                }
////            }catch(Throwable e){
////                e.printStackTrace();
////            }
//        }else{
//            try {
//                RuntimeVariables.sDexLoadBooster.getClass().getDeclaredMethod("setVerificationEnabled", boolean.class).invoke(RuntimeVariables.sDexLoadBooster, false);
//            } catch (Throwable e){
//                e.printStackTrace();
//            }
//        }


        if(!TextUtils.isEmpty(mInstalledVersionName)){
            RuntimeVariables.sInstalledVersionName = mInstalledVersionName;
        }
        System.out.print(SoLoader.class.getName());
        try {
            String preLaunchStr = (String) RuntimeVariables.getFrameworkProperty("preLaunch");
            if (!TextUtils.isEmpty(preLaunchStr)) {
                AtlasPreLauncher launcher = (AtlasPreLauncher) Class.forName(preLaunchStr).newInstance();
                if (launcher != null) {
                    launcher.initBeforeAtlas(mRawApplication.getBaseContext());
                }
            }
        } catch (Throwable e) {
            throw new RuntimeException(e);
        }


        // *2 init atlas use reflect
        boolean multidexEnable = false;
        try {
            ApplicationInfo appInfo = mRawApplication.getPackageManager()
                    .getApplicationInfo(mRawApplication.getPackageName(),
                            PackageManager.GET_META_DATA);
            mRealApplicationName = appInfo.metaData.getString("REAL_APPLICATION");
            multidexEnable = appInfo.metaData.getBoolean("multidex_enable");
        }catch(PackageManager.NameNotFoundException e){
            throw new RuntimeException(e);
        }

        if(multidexEnable){
            MultiDex.install(mRawApplication);
        }

        mRealApplicationName = TextUtils.isEmpty(mRealApplicationName) ? "android.app.Application" : mRealApplicationName;
        if(mRealApplicationName.startsWith(".")){
            mRealApplicationName = mRawApplication.getPackageName() + mRealApplicationName;
        }
        RuntimeVariables.sRealApplicationName = mRealApplicationName;

        try {

            Atlas.getInstance().init(mRawApplication, mIsUpdated);
        } catch (Throwable e) {
                File storageDir = new File(mRawApplication.getFilesDir(),"storage");
                Framework.deleteDirectory(storageDir);
                KernalVersionManager.instance().removeBaseLineInfo();
                android.os.Process.killProcess(Process.myPid());
//                throw new RuntimeException("atlas initialization fail--->" + e.getMessage());
        }
        //////////////////////////////////////////launchTime////////////////////
        try{
            Class BuildConfig = Class.forName(mRawApplication.getPackageName()+".BuildConfig");
            Field launchTimeField = BuildConfig.getDeclaredField("launchTime");
            launchTimeField.setAccessible(true);
            launchTimeField.set(BuildConfig,System.currentTimeMillis());
        }catch(Throwable e){}
        //////////////////////////////////////////launchTime////////////////////

        // *3 remove providerinfo for so installcontentprovider is delayed
        try {
            Object activityThread = AndroidHack.getActivityThread();
            Object mBoundApplication = AtlasHacks.ActivityThread_mBoundApplication.get(activityThread);
            mBoundApplication_provider = AtlasHacks.ActivityThread$AppBindData_providers.get(mBoundApplication);
            if(mBoundApplication_provider!=null && mBoundApplication_provider.size()>0){
                AtlasHacks.ActivityThread$AppBindData_providers.set(mBoundApplication,null);
            }else if (Build.VERSION.SDK_INT >= 24 && mCurrentProcessname!= null && mCurrentProcessname.equals(mRawApplication.getPackageName())){
                mBoundApplication_provider = new ArrayList<>();
                ProviderInfo providerInfo = mRawApplication.getPackageManager().resolveContentProvider(mRawApplication.getPackageName()+".update.provider",0);
                if (providerInfo!= null){
                    providerInfo.exported = false;
                    providerInfo.grantUriPermissions = true;
                    mBoundApplication_provider.add(providerInfo);
                }

            }
        } catch (Exception e) {
            if(e instanceof InvocationTargetException){
                throw new RuntimeException(((InvocationTargetException)e).getTargetException());
            }else {
                throw new RuntimeException(e);
            }
        }
    }

    public void onCreate(){
        try {
            // *3 create real Application
            mRealApplication = (Application) mRawApplication.getBaseContext().getClassLoader().loadClass(mRealApplicationName).newInstance();

            Object activityThread = AndroidHack.getActivityThread();
            //replace baseContext.mOuterContext
            AtlasHacks.ContextImpl_setOuterContext.invoke(mRawApplication.getBaseContext(), mRealApplication);
            //replace baseContext.mPackageInfo.mApplication
            Object mPackageInfo = AtlasHacks.ContextImpl_mPackageInfo.get(mRawApplication.getBaseContext());
            AtlasHacks.LoadedApk_mApplication.set(mPackageInfo,mRealApplication);
            //replace baseContext.mPackageInfo.mActivityThread.mInitialApplication
            AtlasHacks.ActivityThread_mInitialApplication.set(activityThread,mRealApplication);
            //update  baseContext.mPackageInfo.mActivityThread.mAllApplications
            List<Application> allApplications = AtlasHacks.ActivityThread_mAllApplications.get(activityThread);
            for (int i = 0; i < allApplications.size(); i++) {
                if (allApplications.get(i) == mRawApplication) {
                    //替换掉ActivityThread.mAllApplications**
                    allApplications.set(i, mRealApplication);
                }
            }
            RuntimeVariables.androidApplication = mRealApplication;

            /**
             * configuration update
             */
            mRealApplication.registerComponentCallbacks(new ComponentCallbacks() {
                @Override
                public void onConfigurationChanged(Configuration newConfig) {
                    DisplayMetrics newMetrics = new DisplayMetrics();
                    if(RuntimeVariables.delegateResources!=null && RuntimeVariables.androidApplication!=null){
                        WindowManager manager = (WindowManager) RuntimeVariables.androidApplication.getSystemService(Context.WINDOW_SERVICE);
                        if(manager==null || manager.getDefaultDisplay()==null){
                            Log.e("BridgeApplication","get windowmanager service failed");
                            return;
                        }
                        manager.getDefaultDisplay().getMetrics(newMetrics);
                        RuntimeVariables.delegateResources.updateConfiguration(newConfig,newMetrics);
                        try {
                            Method method = Resources.class.getDeclaredMethod("updateSystemConfiguration",
                                    Configuration.class,DisplayMetrics.class,Class.forName("android.content.res.CompatibilityInfo"));
                            method.setAccessible(true);
                            method.invoke(RuntimeVariables.delegateResources,newConfig,newMetrics,null);
                        } catch (Throwable e) {
                            e.printStackTrace();
                        }
                    }
                }

                @Override
                public void onLowMemory() {

                }
            });

            AtlasHacks.Application_attach.invoke(mRealApplication,mRawApplication.getBaseContext());
            // install content providers
            if (mBoundApplication_provider != null && mBoundApplication_provider.size() > 0) {
                Object mBoundApplication = AtlasHacks.ActivityThread_mBoundApplication.get(activityThread);
                AtlasHacks.ActivityThread$AppBindData_providers.set(mBoundApplication,mBoundApplication_provider);
                AtlasHacks.ActivityThread_installContentProviders.invoke(activityThread,mRealApplication,mBoundApplication_provider);
            }else {

            }

        }catch(Throwable e){
            if(e instanceof InvocationTargetException){
                throw new RuntimeException(((InvocationTargetException)e).getTargetException());
            }else {
                throw new RuntimeException(e);
            }
        }

        Atlas.getInstance().startup(mRealApplication,mIsUpdated);

        mRealApplication.onCreate();

    }

}
