/*
 *
 *                                Apache License
 *                          Version 2.0, January 2004
 *                       http://www.apache.org/licenses/
 *
 *  TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *  1. Definitions.
 *
 *     "License" shall mean the terms and conditions for use, reproduction,
 *     and distribution as defined by Sections 1 through 9 of this document.
 *
 *     "Licensor" shall mean the copyright owner or entity authorized by
 *     the copyright owner that is granting the License.
 *
 *     "Legal Entity" shall mean the union of the acting entity and all
 *     other entities that control, are controlled by, or are under common
 *     control with that entity. For the purposes of this definition,
 *     "control" means (i) the power, direct or indirect, to cause the
 *     direction or management of such entity, whether by contract or
 *     otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *     outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *     "You" (or "Your") shall mean an individual or Legal Entity
 *     exercising permissions granted by this License.
 *
 *     "Source" form shall mean the preferred form for making modifications,
 *     including but not limited to software source code, documentation
 *     source, and configuration files.
 *
 *     "Object" form shall mean any form resulting from mechanical
 *     transformation or translation of a Source form, including but
 *     not limited to compiled object code, generated documentation,
 *     and conversions to other media types.
 *
 *     "Work" shall mean the work of authorship, whether in Source or
 *     Object form, made available under the License, as indicated by a
 *     copyright notice that is included in or attached to the work
 *     (an example is provided in the Appendix below).
 *
 *     "Derivative Works" shall mean any work, whether in Source or Object
 *     form, that is based on (or derived from) the Work and for which the
 *     editorial revisions, annotations, elaborations, or other modifications
 *     represent, as a whole, an original work of authorship. For the purposes
 *     of this License, Derivative Works shall not include works that remain
 *     separable from, or merely link (or bind by name) to the interfaces of,
 *     the Work and Derivative Works thereof.
 *
 *     "Contribution" shall mean any work of authorship, including
 *     the original version of the Work and any modifications or additions
 *     to that Work or Derivative Works thereof, that is intentionally
 *     submitted to Licensor for inclusion in the Work by the copyright owner
 *     or by an individual or Legal Entity authorized to submit on behalf of
 *     the copyright owner. For the purposes of this definition, "submitted"
 *     means any form of electronic, verbal, or written communication sent
 *     to the Licensor or its representatives, including but not limited to
 *     communication on electronic mailing lists, source code control systems,
 *     and issue tracking systems that are managed by, or on behalf of, the
 *     Licensor for the purpose of discussing and improving the Work, but
 *     excluding communication that is conspicuously marked or otherwise
 *     designated in writing by the copyright owner as "Not a Contribution."
 *
 *     "Contributor" shall mean Licensor and any individual or Legal Entity
 *     on behalf of whom a Contribution has been received by Licensor and
 *     subsequently incorporated within the Work.
 *
 *  2. Grant of Copyright License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     copyright license to reproduce, prepare Derivative Works of,
 *     publicly display, publicly perform, sublicense, and distribute the
 *     Work and such Derivative Works in Source or Object form.
 *
 *  3. Grant of Patent License. Subject to the terms and conditions of
 *     this License, each Contributor hereby grants to You a perpetual,
 *     worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *     (except as stated in this section) patent license to make, have made,
 *     use, offer to sell, sell, import, and otherwise transfer the Work,
 *     where such license applies only to those patent claims licensable
 *     by such Contributor that are necessarily infringed by their
 *     Contribution(s) alone or by combination of their Contribution(s)
 *     with the Work to which such Contribution(s) was submitted. If You
 *     institute patent litigation against any entity (including a
 *     cross-claim or counterclaim in a lawsuit) alleging that the Work
 *     or a Contribution incorporated within the Work constitutes direct
 *     or contributory patent infringement, then any patent licenses
 *     granted to You under this License for that Work shall terminate
 *     as of the date such litigation is filed.
 *
 *  4. Redistribution. You may reproduce and distribute copies of the
 *     Work or Derivative Works thereof in any medium, with or without
 *     modifications, and in Source or Object form, provided that You
 *     meet the following conditions:
 *
 *     (a) You must give any other recipients of the Work or
 *         Derivative Works a copy of this License; and
 *
 *     (b) You must cause any modified files to carry prominent notices
 *         stating that You changed the files; and
 *
 *     (c) You must retain, in the Source form of any Derivative Works
 *         that You distribute, all copyright, patent, trademark, and
 *         attribution notices from the Source form of the Work,
 *         excluding those notices that do not pertain to any part of
 *         the Derivative Works; and
 *
 *     (d) If the Work includes a "NOTICE" text file as part of its
 *         distribution, then any Derivative Works that You distribute must
 *         include a readable copy of the attribution notices contained
 *         within such NOTICE file, excluding those notices that do not
 *         pertain to any part of the Derivative Works, in at least one
 *         of the following places: within a NOTICE text file distributed
 *         as part of the Derivative Works; within the Source form or
 *         documentation, if provided along with the Derivative Works; or,
 *         within a display generated by the Derivative Works, if and
 *         wherever such third-party notices normally appear. The contents
 *         of the NOTICE file are for informational purposes only and
 *         do not modify the License. You may add Your own attribution
 *         notices within Derivative Works that You distribute, alongside
 *         or as an addendum to the NOTICE text from the Work, provided
 *         that such additional attribution notices cannot be construed
 *         as modifying the License.
 *
 *     You may add Your own copyright statement to Your modifications and
 *     may provide additional or different license terms and conditions
 *     for use, reproduction, or distribution of Your modifications, or
 *     for any such Derivative Works as a whole, provided Your use,
 *     reproduction, and distribution of the Work otherwise complies with
 *     the conditions stated in this License.
 *
 *  5. Submission of Contributions. Unless You explicitly state otherwise,
 *     any Contribution intentionally submitted for inclusion in the Work
 *     by You to the Licensor shall be under the terms and conditions of
 *     this License, without any additional terms or conditions.
 *     Notwithstanding the above, nothing herein shall supersede or modify
 *     the terms of any separate license agreement you may have executed
 *     with Licensor regarding such Contributions.
 *
 *  6. Trademarks. This License does not grant permission to use the trade
 *     names, trademarks, service marks, or product names of the Licensor,
 *     except as required for reasonable and customary use in describing the
 *     origin of the Work and reproducing the content of the NOTICE file.
 *
 *  7. Disclaimer of Warranty. Unless required by applicable law or
 *     agreed to in writing, Licensor provides the Work (and each
 *     Contributor provides its Contributions) on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *     implied, including, without limitation, any warranties or conditions
 *     of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *     PARTICULAR PURPOSE. You are solely responsible for determining the
 *     appropriateness of using or redistributing the Work and assume any
 *     risks associated with Your exercise of permissions under this License.
 *
 *  8. Limitation of Liability. In no event and under no legal theory,
 *     whether in tort (including negligence), contract, or otherwise,
 *     unless required by applicable law (such as deliberate and grossly
 *     negligent acts) or agreed to in writing, shall any Contributor be
 *     liable to You for damages, including any direct, indirect, special,
 *     incidental, or consequential damages of any character arising as a
 *     result of this License or out of the use or inability to use the
 *     Work (including but not limited to damages for loss of goodwill,
 *     work stoppage, computer failure or malfunction, or any and all
 *     other commercial damages or losses), even if such Contributor
 *     has been advised of the possibility of such damages.
 *
 *  9. Accepting Warranty or Additional Liability. While redistributing
 *     the Work or Derivative Works thereof, You may choose to offer,
 *     and charge a fee for, acceptance of support, warranty, indemnity,
 *     or other liability obligations and/or rights consistent with this
 *     License. However, in accepting such obligations, You may act only
 *     on Your own behalf and on Your sole responsibility, not on behalf
 *     of any other Contributor, and only if You agree to indemnify,
 *     defend, and hold each Contributor harmless for any liability
 *     incurred by, or claims asserted against, such Contributor by reason
 *     of your accepting any such warranty or additional liability.
 *
 *  END OF TERMS AND CONDITIONS
 *
 *  APPENDIX: How to apply the Apache License to your work.
 *
 *     To apply the Apache License to your work, attach the following
 *     boilerplate notice, with the fields enclosed by brackets "[]"
 *     replaced with your own identifying information. (Don't include
 *     the brackets!)  The text should be enclosed in the appropriate
 *     comment syntax for the file format. We also recommend that a
 *     file or class name and description of purpose be included on the
 *     same "printed page" as the copyright notice for easier
 *     identification within third-party archives.
 *
 *  Copyright 2016 Alibaba Group
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *
 */

package android.taobao.atlas.startup;

import android.app.ActivityManager;
import android.content.Context;
import android.os.Build;
import android.os.Environment;
import android.os.Process;
import android.taobao.atlas.startup.patch.KernalConstants;
import android.taobao.atlas.startup.patch.KernalFileLock;
import android.text.TextUtils;
import android.util.Log;
import java.io.*;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import static android.os.Environment.MEDIA_MOUNTED;
import static android.os.Environment.MEDIA_UNKNOWN;

/**
 * Created by guanjie on 15/9/10.
 */
public class KernalVersionManager {

    private static KernalVersionManager sBaseInfoManager;
    private final File  BASELINEINFO_DIR ;
    private final File  BASELINEINFO;
    private final File  BASELINEINFO_NEW;
    private String LAST_VERSIONNAME;
    private String LAST_UPDATE_BUNDLES;
    private String LAST_STORAGE_LOCATION;
    private String CURRENT_UPDATE_BUNDLES;
    private String CURRENT_VERSIONAME;
    public String CURRENT_STORAGE_LOCATION;
    private String DEXPATCH_BUNDLES;
    public String DEXPATCH_STORAGE_LOCATION;
    private HashMap<String,String> currentUpdateBundles = new HashMap<String,String>();
    public ConcurrentHashMap<String,Long> dexPatchBundles = new ConcurrentHashMap<>();

    public boolean cachePreVersion = false;

    private static class SingleTonHolder{
        private final static KernalVersionManager INSTANCE = new KernalVersionManager();
    }

    public static KernalVersionManager instance(){
        return SingleTonHolder.INSTANCE;
    }

    @Override
    public String toString(){
        return String.format("%s@%s--%s--dexPatchBundles:%s",
            LAST_VERSIONNAME,CURRENT_VERSIONAME,
            LAST_UPDATE_BUNDLES,
            TextUtils.isEmpty(DEXPATCH_BUNDLES)?"":DEXPATCH_BUNDLES
        );
    }

    public void reset(){
        LAST_VERSIONNAME = "";
        LAST_UPDATE_BUNDLES = "";
        LAST_STORAGE_LOCATION="";
        CURRENT_STORAGE_LOCATION="";
        DEXPATCH_STORAGE_LOCATION="";
        CURRENT_VERSIONAME = "";
        CURRENT_UPDATE_BUNDLES="";
        DEXPATCH_BUNDLES = "";
        currentUpdateBundles.clear();
        dexPatchBundles.clear();
    }

    private KernalVersionManager(){
        BASELINEINFO_DIR = new File(KernalConstants.baseContext.getFilesDir().getAbsolutePath() + File.separatorChar + "bundleBaseline");
        BASELINEINFO     = new File(BASELINEINFO_DIR,"updateInfo");
        BASELINEINFO_NEW = new File(BASELINEINFO_DIR,"updateInfo_new");
    }

    public void init(){
        KernalFileLock.getInstance().LockExclusive(BASELINEINFO_DIR);
        if(shouldSyncUpdateInThisProcess()){
            if(BASELINEINFO_NEW.exists()){
                killChildProcesses(KernalConstants.baseContext);
                int retry  = 3;
                do {
                    if (BASELINEINFO.exists()) {
                        BASELINEINFO.delete();
                    }
                    retry--;
                }while(retry>0 && BASELINEINFO.exists());
                long length = BASELINEINFO_NEW.length();
                retry = 3;
                do {
                    if(BASELINEINFO_NEW.renameTo(BASELINEINFO)){
                        break;
                    }
                    retry--;
                }while(retry>0);
                if(length != BASELINEINFO.length()){
                    throw new RuntimeException("rename baselineinfo fail");
                }
            }
        }
        KernalFileLock.getInstance().unLock(BASELINEINFO_DIR);

        String baselineVersion = "";
        String updateBundles = "";
        String lastVersionName = "";
        String lastUpdateBundles="";
        String lastStorageLocation="";
        String currentStorageLocation="";
        String dexpatchStorageLocation="";
        String dexPatchBundles="";
        if(BASELINEINFO.exists()){
            try {
                /**
                 * 写入顺序
                 * 上一次versionname
                 * 上一次更新内容
                 * 上一次的存储地址
                 * 本次versionname
                 * 本次更新内容
                 * 本次存储地址
                 * dexpatch 更新内容
                 * dexpatch 存储地址
                 * 是否发生了回滚
                 */
                DataInputStream input = new DataInputStream(new BufferedInputStream(new FileInputStream(BASELINEINFO)));
                lastVersionName = input.readUTF();
                lastUpdateBundles = input.readUTF();
                lastStorageLocation=input.readUTF();
                baselineVersion = input.readUTF();
                updateBundles = input.readUTF();
                currentStorageLocation=input.readUTF();
                dexPatchBundles = input.readUTF();
                dexpatchStorageLocation = input.readUTF();
                cachePreVersion = input.readBoolean();
                if(!TextUtils.isEmpty(dexPatchBundles) && !TextUtils.isEmpty(dexpatchStorageLocation)){
                    //external dexpatch
                    File storage = new File(dexpatchStorageLocation);
                    if(!storage.exists() || !getStorageState(storage).equals(Environment.MEDIA_MOUNTED)){
                        dexPatchBundles="";
                        dexpatchStorageLocation = "";
                    }
                }

                if(!TextUtils.isEmpty(updateBundles) && !TextUtils.isEmpty(currentStorageLocation)){
                    //no dexpatch,but external bundle update
                    File storage = new File(currentStorageLocation);
                    if(!storage.exists() || !getStorageState(storage).equals(Environment.MEDIA_MOUNTED)){
                        throw new IOException("update bundle location storage is not usable");
                    }
                }

                try {
                    input.close();
                }catch(Throwable e){}
            } catch (Throwable e) {
                if(KernalConstants.PROCESS.equals(KernalConstants.baseContext.getPackageName())) {
                    rollbackHardly();
                }
                killChildProcesses(KernalConstants.baseContext);
                android.os.Process.killProcess(Process.myPid());
            }
        }
        LAST_VERSIONNAME = lastVersionName;
        LAST_UPDATE_BUNDLES = lastUpdateBundles;
        CURRENT_VERSIONAME = baselineVersion;
        CURRENT_UPDATE_BUNDLES= updateBundles;
        DEXPATCH_BUNDLES = dexPatchBundles;
        LAST_STORAGE_LOCATION = lastStorageLocation;
        CURRENT_STORAGE_LOCATION = currentStorageLocation;
        DEXPATCH_STORAGE_LOCATION = dexpatchStorageLocation;
        parseUpdatedBundles();
    }

    public void removeBaseLineInfo(){
        if(BASELINEINFO_DIR.exists()) {
            deleteDirectory(BASELINEINFO_DIR);
        }
        File bundleupdate = new File(KernalConstants.baseContext.getFilesDir(),"bundleupdate");
        if(bundleupdate.exists()) {
            deleteDirectory(bundleupdate);
        }
    }

    public void deleteDirectory(final File path) {
        final File[] files = path.listFiles();
        if (files == null){
            return;
        }
        for (int i = 0; i < files.length; i++) {
            if (files[i].isDirectory()) {
                deleteDirectory(files[i]);
            } else {
                files[i].delete();
            }
        }
        path.delete();
    }

    public String getBaseBundleVersion(String bundleName){
        return currentUpdateBundles.get(bundleName);
    }

    public long getDexPatchBundleVersion(String bundleName){
        if(dexPatchBundles.containsKey(bundleName)){
            return dexPatchBundles.get(bundleName);
        }else {
            return -1;
        }
    }

    public Set<String> getUpdateBundles(){
        return currentUpdateBundles.keySet();
    }

    public String lastVersionName(){
        return LAST_VERSIONNAME;
    }

    public String currentVersionName(){
        return CURRENT_VERSIONAME;
    }

    public boolean isCachePreVersion(){
        boolean flag = Boolean.FALSE.booleanValue();
        if(flag) {
            Log.e("KernalVersionManager", "can no be inlined");
        }
        return cachePreVersion;
    }

    //updated
    public boolean isUpdated(String bundleName){
        if(!TextUtils.isEmpty(getBaseBundleVersion(bundleName))){
            return true;
        }
        return false;
    }

    //dexpatched
    public boolean isDexPatched(String bundleName){
        return getDexPatchBundleVersion(bundleName)>0;
    }

    public synchronized void parseUpdatedBundles(){
        if(!TextUtils.isEmpty(CURRENT_UPDATE_BUNDLES)){
            String[] bundles = CURRENT_UPDATE_BUNDLES.split(";");
            if(bundles!=null && bundles.length>0){
                for(String bundleInfo : bundles){
                    String[] infoItems = bundleInfo.split("@");
                    currentUpdateBundles.put(infoItems[0],infoItems[1]);
                }
            }
        }
        if(!TextUtils.isEmpty(DEXPATCH_BUNDLES)){
            String[] bundles = DEXPATCH_BUNDLES.split(";");
            if(bundles!=null && bundles.length>0){
                for(String bundleInfo : bundles){
                    String[] infoItems = bundleInfo.split("@");
                    dexPatchBundles.put(infoItems[0],Long.parseLong(infoItems[1]));
                }
            }
        }
    }

    public void rollbackHardly(){
        try {
            File baseLineDir = BASELINEINFO_DIR;
            if (!baseLineDir.exists()) {
                baseLineDir.mkdirs();
            }
            File deprecated = new File(baseLineDir, "deprecated_mark");
            if(!deprecated.exists()) {
                deprecated.createNewFile();
            }
            File file = BASELINEINFO;
            if(file.exists()){
                file.delete();
            }
        }catch(Exception e){
        }
    }

    public void rollback(){
        File baseinfoDir = new File(BASELINEINFO_DIR.getAbsolutePath());
        if (!baseinfoDir.exists()) {
            baseinfoDir.mkdir();
        }
        try {
            /**
             * 写入顺序
             * 上一次versionname
             * 上一次更新内容
             * 本次versionname
             * 本次更新内容
             * dexpatch versionname
             * dexpatch 更新内容
             * 是否发生了回滚
             */
            DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(new File(baseinfoDir.getAbsolutePath(), "baselineInfo"))));

            out.writeUTF("");
            out.writeUTF("");
            out.writeUTF("");
            if(cachePreVersion){
                out.writeUTF(LAST_VERSIONNAME);
                out.writeUTF(LAST_UPDATE_BUNDLES);
                out.writeUTF(LAST_STORAGE_LOCATION);
            }else {
                out.writeUTF("");
                out.writeUTF("");
                out.writeUTF("");
            }
            out.writeUTF("");
            out.writeUTF("");
            out.writeBoolean(cachePreVersion);
            out.flush();
            out.close();
        } catch (Throwable e) {
            e.printStackTrace();
            rollbackHardly();
        }

    }

    public void saveDexPatchInfo(HashMap<String,String> infos,String storageLocation) throws IOException{
        for(Iterator iterator = infos.entrySet().iterator(); iterator.hasNext();)
        {
            Map.Entry entry = (java.util.Map.Entry)iterator.next();
            if(entry.getValue().equals("-1") && dexPatchBundles.containsKey(entry.getKey())){
                //dexPatchBundles.remove(entry.getKey());
            }else{
                dexPatchBundles.put((String)entry.getKey(),Long.parseLong((String)entry.getValue()));
            }
        }

        StringBuilder bundleList = new StringBuilder("");
        for(Iterator iterator = dexPatchBundles.entrySet().iterator(); iterator.hasNext();)
        {
            Map.Entry entry = (java.util.Map.Entry)iterator.next();
            if("-1".equals(infos.get(entry.getKey()))){
                continue;
            }
            bundleList.append(entry.getKey());
            bundleList.append("@");
            bundleList.append(entry.getValue());
            bundleList.append(";");
        }

        File baseinfoFile = BASELINEINFO_DIR;
        if(!baseinfoFile.exists()){
            baseinfoFile.mkdirs();
        }
        File newBaselineInfoFile = new File(BASELINEINFO.getAbsolutePath());
        if(!newBaselineInfoFile.exists()){
            newBaselineInfoFile.createNewFile();
        }

        DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(new File(newBaselineInfoFile.getAbsolutePath()))));
        String bundleListStr = bundleList.toString();
        /**
         * 写入顺序
         * 上一次versionname
         * 上一次更新内容
         * 本次versionname
         * 本次更新内容
         * dexpatch versionname
         * dexpatch 更新内容
         * 是否发生了回滚
         */
        out.writeUTF(LAST_VERSIONNAME);
        out.writeUTF(LAST_UPDATE_BUNDLES != null ? LAST_UPDATE_BUNDLES : "");
        out.writeUTF(LAST_STORAGE_LOCATION!=null ? LAST_STORAGE_LOCATION:"");
        //如果是和基线dexPatch，由于没有做过动态部署，updateinfo文件可能不存在。
        //下次启动时，kernealbundle.patchKernalDex方法在比对版本时versionName为""，启动异常。
        out.writeUTF(TextUtils.isEmpty(CURRENT_VERSIONAME)? KernalConstants.INSTALLED_VERSIONNAME:CURRENT_VERSIONAME);
        out.writeUTF(CURRENT_UPDATE_BUNDLES);
        out.writeUTF(CURRENT_STORAGE_LOCATION);
        //dexpatch 部分
        out.writeUTF(bundleListStr);
        out.writeUTF(storageLocation!=null ? storageLocation : "");
        out.writeBoolean(cachePreVersion);
        out.flush();
        out.close();
    }

    public void saveUpdateInfo(String newBaselineVersion, HashMap<String,String> infos,boolean cachePreVersion,String storageLocation) throws IOException{
        StringBuilder bundleList = new StringBuilder("");
        for(Iterator iterator = infos.entrySet().iterator(); iterator.hasNext();)
        {
            Map.Entry entry = (java.util.Map.Entry)iterator.next();
            bundleList.append(entry.getKey());
            bundleList.append("@");
            bundleList.append(entry.getValue());
            bundleList.append(";");
        }
        File baseinfoFile = BASELINEINFO_DIR;
        if(!baseinfoFile.exists()){
            baseinfoFile.mkdirs();
        }
        File newBaselineInfoFile = new File(BASELINEINFO_NEW.getAbsolutePath());
        if(!newBaselineInfoFile.exists()){
            newBaselineInfoFile.createNewFile();
        }
        DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(new File(newBaselineInfoFile.getAbsolutePath()))));
        String bundleListStr = bundleList.toString();
        /**
         * 写入顺序
         * 上一次versionname
         * 上一次更新内容
         * 上一次的存储地址
         * 本次versionname
         * 本次更新内容
         * 本次存储地址
         * dexpatch 更新内容
         * dexpatch 存储地址
         * 是否发生了回滚
         */
        if(cachePreVersion) {
            out.writeUTF(!TextUtils.isEmpty(CURRENT_VERSIONAME) ? CURRENT_VERSIONAME : "");
            out.writeUTF(CURRENT_UPDATE_BUNDLES != null ? CURRENT_UPDATE_BUNDLES : "");
            out.writeUTF(CURRENT_STORAGE_LOCATION != null ? CURRENT_STORAGE_LOCATION : "");
        }else{
            out.writeUTF("");
            out.writeUTF("");
            out.writeUTF("");
        }
        out.writeUTF(newBaselineVersion);
        out.writeUTF(bundleListStr);
        out.writeUTF(storageLocation!=null ? storageLocation : "");
        //dexpatch 部分
        out.writeUTF("");
        out.writeUTF("");
        out.writeBoolean(cachePreVersion);
        out.flush();
        out.close();
    }

    private boolean shouldSyncUpdateInThisProcess(){
        String processName = KernalConstants.PROCESS;
        if(processName!=null &&
                (processName.equals(KernalConstants.baseContext.getPackageName()) ||
                        processName.toLowerCase().contains(":safemode")
                )){
            return true;
        }else{
            return false;
        }
    }

    public void killChildProcesses(Context context) {
        try {
            long uid = context.getApplicationInfo().uid;
            ActivityManager am = (ActivityManager) context.getSystemService(Context.ACTIVITY_SERVICE);
            List<ActivityManager.RunningAppProcessInfo> a = am.getRunningAppProcesses();
            for (int i = 0; i < a.size(); i++) {
                ActivityManager.RunningAppProcessInfo b = a.get(i);
                if(b.uid == uid && !b.processName.equals(context.getPackageName())){
                    android.os.Process.killProcess(b.pid);
                }

            }
        } catch (Exception e) {

        }
    }

    private String getStorageState(File path) {
        if(path.getAbsolutePath().startsWith(KernalConstants.baseContext.getFilesDir().getAbsolutePath())){
            return MEDIA_MOUNTED;
        }
        final int version = Build.VERSION.SDK_INT;
        if (version >= 19) {
            return Environment.getStorageState(path);
        }

        try {
            final String canonicalPath = path.getCanonicalPath();
            final String canonicalExternal = Environment.getExternalStorageDirectory()
                    .getCanonicalPath();

            if (canonicalPath.startsWith(canonicalExternal)) {
                return Environment.getExternalStorageState();
            }
        } catch (IOException e) {
        }

        return MEDIA_UNKNOWN;
    }

}
