/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.android.build.gradle.internal.pipeline;

import com.android.annotations.NonNull;
import com.android.annotations.Nullable;
import com.android.build.api.transform.QualifiedContent;
import com.android.build.api.transform.QualifiedContent.Scope;
import com.android.build.api.transform.Transform;
import com.android.build.gradle.internal.scope.VariantScope;
import com.android.build.gradle.internal.tasks.DefaultAndroidTask;
import com.android.builder.model.AndroidProject;
import com.android.builder.profile.ThreadRecorder;
import com.android.utils.FileUtils;
import com.android.utils.StringHelper;
import com.google.common.collect.Lists;

import org.apache.commons.lang3.reflect.FieldUtils;
import org.gradle.api.Project;
import org.gradle.api.tasks.StopExecutionException;
import org.gradle.api.tasks.TaskCollection;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.SortedMap;

import static com.android.utils.StringHelper.capitalize;

/**
 * Inject the injection management class with TransformManger, Mainly injected into the intermediate process of class->dex
 * Created by shenghua.nish on 2016-06-13 2:02 PM.
 */
public class InjectTransformManager {

    private static final String FD_TRANSFORMS = "transforms";

    private List<Transform> transforms = Lists.newArrayList();

    private List<String> transformTaskList = Lists.newArrayList();

    private final Project project;

    private final String variantName;

    public InjectTransformManager(Project project, @NonNull String variantName) {
        this.project = project;
        this.variantName = variantName;
    }

    public void reset() {
        transforms.clear();
        transformTaskList.clear();
    }

    /**
     * Add a Transform operation before some Transform
     *
     * @param transformClazz
     * @param injectTransform
     * @param scope
     * @param <T>
     * @return
     */
    public <T extends InjectTransform> TransformTask addInjectTransformBeforeTransform(Class<? extends Transform> transformClazz,
                                                                                       T injectTransform,
                                                                                       @NonNull VariantScope scope) {
        TaskCollection<DefaultAndroidTask> androidTasks = project.getTasks()
                .withType(DefaultAndroidTask.class);
        SortedMap<String, DefaultAndroidTask> androidTaskSortedMap = androidTasks.getAsMap();
        TransformTask oprTransformTask = null; // The task to be inserted
        for (String taskName : androidTaskSortedMap.keySet()) {
            DefaultAndroidTask androidTask = androidTaskSortedMap.get(taskName);
            if (variantName.equals(androidTask.getVariantName())) {
                if (androidTask instanceof TransformTask &&
                        ((TransformTask) androidTask).getTransform()
                                .getClass()
                                .equals(transformClazz)) {
                    oprTransformTask = (TransformTask) androidTask;
                    break;
                }
            }
        }
        if (null == oprTransformTask) {
            throw new StopExecutionException("TransformTask with transfrom type:" +
                                                     transformClazz.getName() +
                                                     " can not found!");
        }
        transforms.add(injectTransform);
        //Determines whether the two Transform dependencies are correct, that is, the output type is consistent with the next input type
        checkTransformConfig(oprTransformTask.getTransform(), injectTransform);

        String taskName = scope.getTaskName(getTaskNamePrefix(injectTransform));

        try {
            IntermediateStream outputStream = getOutputStream(injectTransform, scope, taskName);

            // Configuration TransformTask
            TransformTaskParam transformTaskParam = getTransformParam(oprTransformTask);
            TransformTask.ConfigAction<T> configAction = new TransformTask.ConfigAction<T>(
                    variantName,
                    taskName,
                    injectTransform,
                    transformTaskParam.consumedInputStreams,
                    transformTaskParam.referencedInputStreams,
                    outputStream,
                    ThreadRecorder.get(),null);

            TransformTask injectTransformTask = project.getTasks()
                    .create(configAction.getName(), configAction.getType());
            oprTransformTask.dependsOn(injectTransformTask);
            for (TransformStream transformStream : transformTaskParam.consumedInputStreams) {
                if (null != transformStream.getFileCollection()) {
                    injectTransformTask.dependsOn(transformStream.getFileCollection());
                }
            }

//            for (TransformStream transformStream : transformTaskParam.referencedInputStreams) {
//                if (null != transformStream.getDependencies()) {
//                    injectTransformTask.dependsOn(transformStream.getDependencies());
//                }
//            }

            if (transformTaskList.size() > 0) {
                injectTransformTask.dependsOn(transformTaskList.get(transformTaskList.size() - 1));
            }
            transformTaskList.add(taskName);
            configAction.execute(injectTransformTask);

            //Modify the input of the oprTransformTask
            if (injectTransform.updateNextTransformInput()) {
                Collection<TransformStream> newInputStream = Lists.newArrayList();
                newInputStream.add(outputStream);
                updateTransformTaskConfig(oprTransformTask,
                                          newInputStream,
                                          transformTaskParam.referencedInputStreams,
                                          transformTaskParam.outputStream);
            }

            return injectTransformTask;
        } catch (IllegalAccessException e) {
            throw new StopExecutionException(e.getMessage());
        }
    }

    /**
     * Get the task outputStream
     *
     * @param injectTransform
     * @param scope
     * @param taskName
     * @param <T>
     * @return
     */
    @NotNull
    private <T extends Transform> IntermediateStream getOutputStream(T injectTransform,
                                                                     @NonNull VariantScope scope,
                                                                     String taskName) {
        File outRootFolder = FileUtils.join(project.getBuildDir(),
                                            StringHelper.toStrings(AndroidProject.FD_INTERMEDIATES,
                                                                   FD_TRANSFORMS,
                                                                   injectTransform.getName(),
                                                                   scope.getDirectorySegments()));

        Set<? super Scope> requestedScopes = injectTransform.getScopes();

        // create the output
        return IntermediateStream.builder(project,injectTransform.getName())
                .addContentTypes(injectTransform.getOutputTypes())
                .addScopes(requestedScopes)
                .setTaskName(taskName)
                .setRootLocation(outRootFolder).build();

    }

    /**
     * Gets the list of all transformtasks
     *
     * @return
     */
    private SortedMap<String, TransformTask> getTransformTasks() {
        TaskCollection<TransformTask> transformTasks = project.getTasks()
                .withType(TransformTask.class);
        return transformTasks.getAsMap();
    }

    /**
     * Update the parameters for the transformTask
     *
     * @param transformTask
     * @param consumedInputStreams
     * @param referencedInputStreams
     * @param outputStream
     */
    private void updateTransformTaskConfig(TransformTask transformTask,
                                           @NonNull Collection<TransformStream> consumedInputStreams,
                                           @NonNull Collection<TransformStream> referencedInputStreams,
                                           @Nullable IntermediateStream outputStream) throws IllegalAccessException {
        Field consumedInputStreamsField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                                      "consumedInputStreams",
                                                                      true);
        Field referencedInputStreamsField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                                        "referencedInputStreams",
                                                                        true);
        Field outputStreamField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                              "outputStream",
                                                              true);

        if (null == consumedInputStreamsField ||
                null == referencedInputStreamsField ||
                null == outputStreamField) {
            throw new StopExecutionException(
                    "The TransformTask does not has field with name: consumedInputStreams or referencedInputStreams or outputStream! Plugin version does not support!");
        }
        consumedInputStreamsField.set(transformTask, consumedInputStreams);
        referencedInputStreamsField.set(transformTask, referencedInputStreams);
        outputStreamField.set(transformTask, outputStream);
    }

    /**
     * Gets the parameters of a transformTask
     *
     * @param transformTask
     * @return
     * @throws IllegalAccessException
     */
    private TransformTaskParam getTransformParam(TransformTask transformTask) throws IllegalAccessException {
        TransformTaskParam transformTaskParam = new TransformTaskParam();
        Field consumedInputStreamsField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                                      "consumedInputStreams",
                                                                      true);
        Field referencedInputStreamsField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                                        "referencedInputStreams",
                                                                        true);
        Field outputStreamField = FieldUtils.getDeclaredField(StreamBasedTask.class,
                                                              "outputStream",
                                                              true);

        if (null == consumedInputStreamsField ||
                null == referencedInputStreamsField ||
                null == outputStreamField) {
            throw new StopExecutionException(
                    "The TransformTask does not has field with name: consumedInputStreams or referencedInputStreams or outputStream! Plugin version does not support!");
        }
        transformTaskParam.consumedInputStreams = (Collection<TransformStream>) consumedInputStreamsField
                .get(transformTask);
        transformTaskParam.referencedInputStreams = (Collection<TransformStream>) referencedInputStreamsField
                .get(transformTask);
        transformTaskParam.outputStream = (IntermediateStream) outputStreamField.get(transformTask);
        return transformTaskParam;
    }

    /**
     * The input type of the injection is consistent with the input type of the next task
     *
     * @param oprTransform
     * @param injectTransform
     */
    private void checkTransformConfig(Transform oprTransform, Transform injectTransform) {
        assert (oprTransform.getInputTypes().equals(injectTransform.getOutputTypes()));
    }

    /**
     * TransformTaskParameter combination
     */
    static class TransformTaskParam {

        Collection<TransformStream> consumedInputStreams;

        Collection<TransformStream> referencedInputStreams;

        IntermediateStream outputStream;
    }

    @NonNull
    private static String getTaskNamePrefix(@NonNull Transform transform) {
        StringBuilder sb = new StringBuilder(100);
        sb.append("transform");

        Iterator<QualifiedContent.ContentType> iterator = transform.getInputTypes().iterator();
        // there's always at least one
        sb.append(capitalize(iterator.next().name().toLowerCase(Locale.getDefault())));
        while (iterator.hasNext()) {
            sb.append("And")
                    .append(capitalize(iterator.next().name().toLowerCase(Locale.getDefault())));
        }

        sb.append("With").append(capitalize(transform.getName())).append("For");

        return sb.toString();
    }
}
