/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.android.builder.core;

import com.android.annotations.NonNull;
import com.android.annotations.Nullable;
import com.android.build.gradle.internal.aapt.AaptGeneration;
import com.android.build.gradle.options.ProjectOptions;
import com.android.builder.internal.aapt.Aapt;
import com.android.builder.internal.aapt.AaptPackageConfig;
import com.android.builder.internal.aapt.AaptPackageConfig.Builder;
import com.android.builder.sdk.SdkInfo;
import com.android.builder.symbols.*;
import com.android.dex.Dex;
import com.android.dx.command.dexer.DxContext;
import com.android.dx.merge.CollisionPolicy;
import com.android.dx.merge.DexMerger;
import com.android.ide.common.process.*;
import com.android.manifmerger.ManifestMerger2;
import com.android.manifmerger.ManifestProvider;
import com.android.manifmerger.MergingReport;
import com.android.sdklib.BuildToolInfo;
import com.android.sdklib.BuildToolInfo.PathId;
import com.android.utils.ILogger;
import com.google.common.collect.Table;
import com.taobao.android.AaptLib;
import com.taobao.android.builder.AtlasBuildContext;
import com.taobao.android.builder.extension.AtlasExtension;
import com.taobao.android.builder.tasks.app.bundle.AtlasSymbolIo;
import com.taobao.android.builder.tools.*;
import com.taobao.android.builder.tools.cache.FileCacheCenter;
import com.taobao.android.builder.tools.cache.FileCacheException;
import com.taobao.android.builder.tools.zip.ZipUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.StringUtils;
import org.gradle.api.GradleException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.*;

/**
 * 1. Custom Android BuilderTool classes that support custom aapt operations
 * 2. Cache optimization for dex , Adjust the dexOptions parameter
 */
public class AtlasBuilder extends AndroidBuilder {

    private static final Logger sLogger = LoggerFactory.getLogger(AtlasBuilder.class);

    public static final String PRE_DEXCACHE_TYPE = "pre-dex-0.11";

    protected AtlasExtension atlasExtension;

    public MultiDexer multiDexer;

    protected AndroidBuilder defaultBuilder;

    private DexByteCodeConverter dexByteCodeConverter;

    private final JavaProcessExecutor javaProcessExecutor;

    public List<ManifestProvider>manifestProviders;

    private final ILogger logger;

    private final boolean verboseExec;
    private boolean useMyDex;

    private AaptGeneration aaptGeneration;

    /**
     * Creates an AndroidBuilder.
     * <p/>
     * <var>verboseExec</var> is needed on top of the ILogger due to remote exec tools not being
     * able to output info and verbose messages separately.
     *
     * @param projectId
     * @param createdBy           the createdBy String for the apk manifest.
     * @param processExecutor
     * @param javaProcessExecutor
     * @param errorReporter
     * @param logger              the Logger
     * @param verboseExec         whether external tools are launched in verbose mode
     */
    public AtlasBuilder(@NonNull String projectId,
                        @Nullable String createdBy,
                        @NonNull ProcessExecutor processExecutor,
                        @NonNull JavaProcessExecutor javaProcessExecutor,
                        @NonNull ErrorReporter errorReporter,
                        @NonNull ILogger logger,
                        boolean verboseExec) {
        super(projectId,
              createdBy,
              processExecutor,
              javaProcessExecutor,
              errorReporter,
              logger,
              verboseExec);
        this.javaProcessExecutor = javaProcessExecutor;
        this.verboseExec = verboseExec;
        this.logger = logger;
    }

    public void initAapt(ProjectOptions projectOptions){
        super.setTargetInfo(defaultBuilder.getTargetInfo());
        super.setSdkInfo(defaultBuilder.getSdkInfo());
        super.setLibraryRequests((Collection<LibraryRequest>) ReflectUtils.getField(defaultBuilder,"mLibraryRequests"));
        boolean updateAapt = atlasExtension.getTBuildConfig().getUseCustomAapt();
        aaptGeneration = AaptGeneration.fromProjectOptions(projectOptions);
        switch (aaptGeneration){
            case AAPT_V1:
                if (updateAapt){
                    updateAapt(PathId.AAPT);
                }
                break;
            // case AAPT_V2:
            //     if (updateAapt){
            //         updateAapt(PathId.AAPT2);
            //     }
            //     break;
            // case AAPT_V2_DAEMON_MODE:
            //     if (updateAapt){
            //         updateAapt(PathId.DAEMON_AAPT2);
            //     }
            //     break;
            //
            // case AAPT_V2_JNI:
            //     if (updateAapt) {
            //         updateAapt(PathId.DAEMON_AAPT2);
            //     }
            //     break;
                default:break;
        }

    }

    public AtlasExtension getAtlasExtension() {
        return atlasExtension;
    }

    public void setAtlasExtension(AtlasExtension atlasExtension) {
        this.atlasExtension = atlasExtension;
    }

    public AndroidBuilder getDefaultBuilder() {
        return defaultBuilder;
    }

    public void setDefaultBuilder(AndroidBuilder defaultBuilder) {
        this.defaultBuilder = defaultBuilder;
    }

    @Override
    public void processResources(Aapt aapt,
                                 Builder aaptConfigBuilder)
        throws IOException, InterruptedException, ProcessException {
        String s;

        if (aaptGeneration == AaptGeneration.AAPT_V1) {
            s = "--non-constant-id";

        } else {
            s = "--non-final-ids";
        }

        if (!atlasExtension.getTBuildConfig().getAaptConstantId() && !aaptConfigBuilder.build()
            .getOptions()
            .getAdditionalParameters()
            .contains(s)) {
            aaptConfigBuilder.build().getOptions().getAdditionalParameters().add(s);
        } else {
            while (aaptConfigBuilder.build().getOptions().getAdditionalParameters().contains(s)){
                aaptConfigBuilder.build().getOptions().getAdditionalParameters().remove(s);
            }
        }

        super.processResources(aapt, aaptConfigBuilder);


        File symboFile = new File(aaptConfigBuilder.build().getSymbolOutputDir(),"R.txt");
        if (!symboFile.exists()){
            throw new ProcessException(symboFile.getAbsolutePath() + "is not exist!");
        }


    }




    /**
     * Deal with awb resources
     *  @param aaptCommand
     * @param enforceUniquePackageName
     * @param processOutputHandler
     * @param mainSymbolFile
     */
    public void processAwbResources(Aapt aapt,AaptPackageConfig.Builder aaptConfigBuilder,
                                    File mainSymbolFile,File awbSymbolFile,String pacakgeName) throws IOException, ProcessException {
        aaptConfigBuilder.setBuildToolInfo(getTargetInfo().getBuildTools());
        aaptConfigBuilder.setAndroidTarget(getTargetInfo().getTarget());
        aaptConfigBuilder.setLogger(logger);
            logger.info("Aapt output file {}");
            AaptPackageConfig aaptConfig = aaptConfigBuilder.build();

        try {
            aapt.link(aaptConfig).get();
        } catch (Exception e) {
            throw new ProcessException("Failed to execute aapt", e);
        }
        File sourceOut = aaptConfig.getSourceOutputDir();
        if (sourceOut != null) {
            // Figure out what the main symbol file's package is.
            String mainPackageName = aaptConfig.getCustomPackageForR();
            if (mainPackageName == null) {
                mainPackageName =
                        SymbolUtils.getPackageNameFromManifest(aaptConfig.getManifestFile());
            }

            // Load the main symbol file.
            File mainRTxt = new File(aaptConfig.getSymbolOutputDir(), "R.txt");
            File mergedSymbolFile = new File(aaptConfig.getSymbolOutputDir(), "R-all.txt");
            try {
                sLogger.info("awbSymbolFile:" + mainRTxt);
                if (null != mainRTxt && mainRTxt.exists()) {
                    FileUtils.copyFile(mainRTxt, mergedSymbolFile);
                }else {
                    throw new ProcessException(mainRTxt.getAbsolutePath() + "is not exist, maybe no resources in this bundle! ");
                }

                SymbolTable awbSymbols = build(awbSymbolFile,mainPackageName,mainRTxt,mainSymbolFile,pacakgeName);
                if (awbSymbols!= null) {
                    AtlasSymbolIo.write(awbSymbols, mergedSymbolFile);
                }

                //why do this? youku need use this !
//                FileUtils.writeLines(mergedSymbolFile, FileUtils.readLines(mainSymbolFile), true);
            } catch (IOException e) {
                throw new RuntimeException("Could not load file ", e);
            }

            SymbolTable mainSymbols =
                    mergedSymbolFile.isFile()
                            ? AtlasSymbolIo.readFromAapt(mergedSymbolFile, mainPackageName)
                            : SymbolTable.builder().tablePackage(mainPackageName).build();


            // For each dependency, load its symbol file.
            Set<SymbolTable> depSymbolTables =
                    SymbolUtils.loadDependenciesSymbolTables(
                            aaptConfig.getLibrarySymbolTableFiles(), mainPackageName);

            boolean finalIds = true;
            if (aaptConfig.getVariantType() == VariantType.LIBRARY) {
                finalIds = false;
            }


            SymbolIo.exportToJava(mainSymbols, sourceOut, finalIds);

            RGeneration.generateRForLibraries(mainSymbols, depSymbolTables, sourceOut, finalIds);
        }
    }



    /**
     * Get the custom buildToolInfo
     *
     * @param targetInfo
     * @return
     */


    public void updateAapt(PathId pathId) {

        if (null == defaultBuilder.getTargetInfo()) {
            return;
            }
            BuildToolInfo defaultBuildToolInfo = defaultBuilder.getTargetInfo().getBuildTools();
            File customAaptFile = getAapt();
            try {
                Method method = defaultBuildToolInfo.getClass()
                    .getDeclaredMethod("add", PathId.class, File.class);
                method.setAccessible(true);
                method.invoke(defaultBuildToolInfo, pathId, customAaptFile);
            } catch (Throwable e) {
                throw new GradleException(e.getMessage());
            }
     }

    @Override
    @NonNull
    public List<String> getBootClasspathAsStrings(boolean includeOptionalLibraries) {

        return defaultBuilder.getBootClasspathAsStrings(includeOptionalLibraries);

    }


    private File getAapt() {

        String osName = "mac";
        String fileName = "aapt";

        if (StringUtils.containsIgnoreCase(System.getProperty("os.name"), "Mac")) {
            osName = "mac";
        } else if (StringUtils.containsIgnoreCase(System.getProperty("os.name"), "Linux")) {
            osName = "linux";
        } else if (StringUtils.containsIgnoreCase(System.getProperty("os.name"), "windows")) {
            osName = "win";
            fileName = "aapt.exe";
        }
        String aaptPath = "aapt/" + osName + "/" + fileName;
        File aaptFile = new File(AaptLib.class.getClassLoader().getResource(aaptPath).getFile());
        if (aaptFile.isFile()) {
            return aaptFile;
        }

        File jarFile = PathUtil.getJarFile(AaptLib.class);
        File jarFolder = new File(jarFile.getParentFile(),
                                  FilenameUtils.getBaseName(jarFile.getName()));
        jarFolder.mkdirs();
        aaptFile = new File(jarFolder, fileName);

        if (!aaptFile.exists()) {
            aaptFile = ZipUtils.extractZipFileToFolder(jarFile, aaptPath,
                                                       aaptFile.getParentFile());
            aaptFile.setExecutable(true);
        }
        return aaptFile;
    }

    //dexMain entrance
    public void convertByteCode(Collection<File> inputs,
                                File outDexFolder,
                                boolean multidex,
                                File mainDexList,
                                DexOptions dexOptions,
                                ProcessOutputHandler processOutputHandler, boolean awb)
        throws IOException, InterruptedException, ProcessException {

        boolean fastMultiDex = false;

        Profiler.start();

        if (atlasExtension.getTBuildConfig().isDexCacheEnabled() && inputs.size() > 0) {

            Profiler.enter("jar2dex");

            List<Dex> dexs = new ArrayList<>();
            Map<File, Dex> fileDexMap = new HashMap<>();

            //Do dexMerge
            File tmpDir = new File(outDexFolder, "tmp");
            tmpDir.mkdirs();

            long startTime = System.currentTimeMillis();

            List<File> outputs = new ArrayList<>();

                for (File input : inputs) {
                    logger.warning("Dex input File:%s",input.getAbsolutePath());

                    final File dexDir = getDexOutputDir(input, tmpDir, outputs);
                    dexDir.mkdirs();

                    preDexLibrary(input, dexDir, multidex, dexOptions, processOutputHandler);

                    outputs.add(dexDir);
            }

                for (File dexDir : outputs) {
                Collection<File> files = FileUtils.listFiles(dexDir,new String[]{"dex"},true);
                for (File dexFile : files) {
                    if (dexFile.exists() && dexFile.length() > 0) {
                        Dex dex = new Dex(dexFile);
                        dexs.add(dex);
                        fileDexMap.put(dexFile, dex);
                    }
                }
            }

            Profiler.release();
            long endDexTime = System.currentTimeMillis();

            Profiler.enter("dexmerge");
            if (dexs.size() > 0) {

                    if (multidex && !awb) {

                        throw new GradleException("can't support~");

                    } else {

                        DexMerger dexMerger = new DexMerger(dexs.toArray(new Dex[0]),
                                                            CollisionPolicy.KEEP_FIRST, new DxContext());
                        Dex dex = dexMerger.merge();

                        File dexFile = new File(outDexFolder, "classes.dex");

                        dex.writeTo(dexFile);

                        if (!dexFile.exists()) {
                            sLogger.error("dexmerge failed, not dex file found , inputs : " + dexs.size());
                        }
                    }



            } else {
                sLogger.error("no dex found to  " + outDexFolder.getAbsolutePath());
            }
            Profiler.release();

            FileUtils.deleteDirectory(tmpDir);

            long finishTime = System.currentTimeMillis();

            sLogger.info("[mtldex] dex consume {} , dexmerge consume {}",
                         endDexTime - startTime,
                         finishTime - endDexTime);
        }
        Profiler.release();
        if (fastMultiDex) {
            sLogger.error("[mtldex] main dex consume {} ", Profiler.dump());
        }

    }

    private File getDexOutputDir(File input, File rootDir, List<File> outputs) {
        File dexDir = null;
        if (input.isFile() && input.getName().endsWith("jar")) {
            dexDir = new File(rootDir, FileNameUtils.getUniqueJarName(input));
        } else {
            dexDir = new File(rootDir, "dexDir" + outputs.size());
        }
        return dexDir;
    }



    public void preDexLibrary(@NonNull File inputFile,
                              @NonNull File outFile,
                              boolean multiDex,
                              @NonNull DexOptions dexOptions,
                              @NonNull ProcessOutputHandler processOutputHandler)
        throws IOException, InterruptedException, ProcessException {

        if (!atlasExtension.getTBuildConfig().isDexCacheEnabled() || multiDex) {
            super.preDexLibrary(inputFile, outFile, multiDex, dexOptions, processOutputHandler,14);
            return;
        }

        String md5 = "";
        File dexFile = new File(outFile, "classes.dex");

        if (!inputFile.getName().startsWith("combined") && !(inputFile.getName().startsWith("main") && inputFile
            .getName().endsWith("jar")) && inputFile.isFile()) {

            if (inputFile.isFile()) {
                md5 = MD5Util.getFileMD5(inputFile);
            } else if (inputFile.isDirectory()) {
                md5 = MD5Util.getFileMd5(FileUtils.listFiles(inputFile,
                                                             new String[] {"class"},
                                                             true));
            }

            //TODO md5 with other
            if (StringUtils.isNotEmpty(md5)) {

                String other = "" + dexOptions.getAdditionalParameters().contains("--useMyDex") + dexOptions
                    .getJumboMode() + dexOptions.getKeepRuntimeAnnotatedClasses();

                try {
                    md5 = MD5Util.getMD5(other + md5);
                } catch (Exception e) {
                    throw new GradleException(e.getMessage(), e);
                }

                try {
                    FileCacheCenter.fetchFile(PRE_DEXCACHE_TYPE, md5, false,
                                              atlasExtension.getTBuildConfig().isDexNetworkCacheEnabled(), dexFile);
                } catch (FileCacheException e) {
                    sLogger.error(e.getMessage(), e);
                }

                if (dexFile.exists() && dexFile.length() > 0) {
                    sLogger.info("[mtldex] hit cache dex for {} , {}",
                                 inputFile.getAbsolutePath(),
                                 dexFile.getAbsolutePath());
                    return;
                } else {
                    sLogger.info("[mtldex] discache dex for {}", inputFile.getAbsolutePath());
                }
            }
        } else {
            //R Too much, you need to start multi-dex
        }

        multiDex = false;

        dexFile.delete();

        //todo  Set the dexOptions
        DefaultDexOptions defaultDexOptions = DefaultDexOptions.copyOf(dexOptions);

        if (!multiDex) {
            defaultDexOptions.setJavaMaxHeapSize("500m");
            defaultDexOptions.setDexInProcess(true);
            defaultDexOptions.getAdditionalParameters().remove("--multi-dex");
        }

        sLogger.info("[mtldex] pre dex for {} {}",
                     inputFile.getAbsolutePath(),
                     outFile.getAbsolutePath());

        super.preDexLibrary(inputFile, outFile, multiDex, defaultDexOptions, processOutputHandler,14);

        if (multiDex) {
            return;
        }

        if (!dexFile.exists() || dexFile.length() == 0) {
            sLogger.warn("dex failed " + dexFile.getAbsolutePath() + "->" + inputFile.getAbsolutePath());
            return;
        }

        sLogger.warn("dex success " + dexFile.getAbsolutePath() + "->" + inputFile.getAbsolutePath());

        if (StringUtils.isNotEmpty(md5) && dexFile.exists()) {
            try {
                FileCacheCenter.cacheFile(PRE_DEXCACHE_TYPE, md5, dexFile,
                                          AtlasBuildContext.sBuilderAdapter.pushCacheToNetwork && atlasExtension
                                              .getTBuildConfig().isDexNetworkCacheEnabled());
            } catch (FileCacheException e) {
                sLogger.error(e.getMessage(), e);
            }

        }
    }

    @Override
    public SdkInfo getSdkInfo() {
        return defaultBuilder.getSdkInfo();
    }

    @Override
    public MergingReport mergeManifestsForApplication(
            @NonNull File mainManifest,
            @NonNull List<File> manifestOverlays,
            @NonNull List<? extends ManifestProvider> dependencies,
            @Nullable String featureName,
            String packageOverride,
            int versionCode,
            String versionName,
            @Nullable String minSdkVersion,
            @Nullable String targetSdkVersion,
            @Nullable Integer maxSdkVersion,
            @NonNull String outManifestLocation,
            @Nullable String outAaptSafeManifestLocation,
            @Nullable String outInstantRunManifestLocation,
            ManifestMerger2.MergeType mergeType,
            Map<String, Object> placeHolders,
            @NonNull List<ManifestMerger2.Invoker.Feature> optionalFeatures,
            @Nullable File reportFile) {


        return super.mergeManifestsForApplication(
                mainManifest,
                manifestOverlays,
                manifestProviders,
                featureName,
                packageOverride,
                versionCode,
                versionName,
                minSdkVersion,
                targetSdkVersion,
                maxSdkVersion,
                outManifestLocation,
                outAaptSafeManifestLocation,
                outInstantRunManifestLocation,
                mergeType,
                placeHolders,
                optionalFeatures,
                reportFile);

    }


    public static interface MultiDexer {

        public Collection<File> repackageJarList(Collection<File> files, File mainDexListFile, boolean release) throws IOException;

        public void dexMerge(Map<File, Dex> fileDexMap, File outDexFolder) throws IOException;

    }


    private SymbolTable build(File awbSymbolFile, String mainPackageName,File mainRTxt,File mainSymbolFile,String pacakgeName) throws IOException {
        SymbolTable.Builder builder = new SymbolTable.Builder();
        SymbolTable awbSymbols = null;
        if (awbSymbolFile.exists()){
            awbSymbols = AtlasSymbolIo.readFromAapt(awbSymbolFile, mainPackageName);
            awbSymbols.getSymbols().values().forEach(symbol -> builder.add(symbol));
        };

        if (mainRTxt.exists()) {
            SymbolTable generateAwbSymbols = AtlasSymbolIo.readFromAapt(mainRTxt, mainPackageName);
            generateAwbSymbols.getSymbols().values().forEach(symbol -> {
                if (builder.contains(symbol)) {
                    removeSymbol(builder, symbol);
                }
                builder.add(symbol);

            });

        }

        SymbolTable mainSymbols = AtlasSymbolIo.readFromAapt(mainSymbolFile, pacakgeName);
        mainSymbols.getSymbols().values().forEach(symbol -> {
            if (builder.contains(symbol)){
                removeSymbol(builder,symbol);
                builder.add(symbol);
            }
        });


        return builder.build();
    }

    private void removeSymbol(SymbolTable.Builder builder, Symbol symbol) {
        Table table = (Table) ReflectUtils.getField(builder,"symbols");
        table.remove(symbol.getResourceType(),symbol.getName());
    }




    public void writeLines(File file,List<String>lines,boolean append){
         Set<String>mergeLines = new LinkedHashSet<>();
        try {
            List<String> readLines = FileUtils.readLines(file);
            mergeLines.addAll(readLines);
            mergeLines.addAll(lines);
            FileUtils.writeLines(file,mergeLines,append);
        } catch (IOException e) {
            e.printStackTrace();
        }


    }




}
