/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.extension;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.google.common.collect.Sets;
import com.taobao.android.builder.extension.annotation.Config;

/**
 * Created by shenghua.nish on 2016-05-17 At 10:15.
 */
public class TBuildConfig {

    @Config(message = "Remote bundle list, artifactId", advance = false, order = 1, group = "atlas")
    private Set<String> outOfApkBundles = Sets.newHashSet();

    public Set<String> getBundleToMdex() {
        return bundleToMdex;
    }

    public void setBundleToMdex(Set<String> bundleToMdex) {
        this.bundleToMdex = bundleToMdex;
    }

    private Set<String>bundleToMdex = Sets.newHashSet();


    public Boolean getAllBundlesToMdex() {
        return allBundlesToMdex;
    }

    public void setAllBundlesToMdex(Boolean allBundlesToMdex) {
        this.allBundlesToMdex = allBundlesToMdex;
    }

    private Boolean allBundlesToMdex = false;

    public Boolean getConsumerProguardEnabled() {
        return consumerProguardEnabled;
    }

    public void setConsumerProguardEnabled(Boolean consumerProguardEnabled) {
        this.consumerProguardEnabled = consumerProguardEnabled;
    }

    private Boolean consumerProguardEnabled = false;

    public Boolean getMergeBundlesDex() {
        return mergeBundlesDex;
    }

    public void setMergeBundlesDex(Boolean mergeBundlesDex) {
        this.mergeBundlesDex = mergeBundlesDex;
    }

    private Boolean mergeBundlesDex = false;

    public Boolean getScanDupRes() {
        return scanDupRes;
    }

    public void setScanDupRes(Boolean scanDupRes) {
        this.scanDupRes = scanDupRes;
    }

    private Boolean scanDupRes = false;

    public String getAppCoordinate() {
        return appCoordinate;
    }

    public void setAppCoordinate(String appCoordinate) {
        this.appCoordinate = appCoordinate;
    }

    @Config(message = "appCoordinate groupId:artifactId", advance = false, order = 1, group = "atlas")
    private String appCoordinate = "";

    public String getGroup() {
        return group;
    }

    public void setGroup(String group) {
        this.group = group;
    }

    @Config(message = "group", advance = false, order = 1, group = "atlas")

    private String group = "";


    @Config(title = "Self-initiated bundle list", message = "The value is packageName", order = 1, advance = false, group = "atlas")
    private List<String> autoStartBundles = new ArrayList<String>();


    @Config(title = "Pre-launch list",
        message = "To implement the PreLaunch class, multiple classes are used , No separate", order = 1, advance = false, group = "atlas")
    private String preLaunch = "";

    @Config(title = "atlas The subcontract",
        message = "atlasThe primary dex subcontracting mechanism, the first dex to set up only the startup code corresponding to atlas", order = 3, advance = false, group = "atlas")
    private boolean atlasMultiDex = false;


    public boolean isUpdateSdkVersion() {
        return updateSdkVersion;
    }

    public void setUpdateSdkVersion(boolean updateSdkVersion) {
        this.updateSdkVersion = updateSdkVersion;
    }

    private boolean updateSdkVersion = true;

    @Config(message = "List of so files to delete", order = 4, advance = true, group = "atlas")
    private Set<String> removeSoFiles = Sets.newHashSet();

    @Config(title = "bundleThe packageId definition file", message = "bundlePackageId defines files that are not defined automatically", group = "atlas")
    private File packageIdFile = new File("");


    public File getAwbConfigFile() {
        return awbConfigFile;
    }

    public void setAwbConfigFile(File awbConfigFile) {
        this.awbConfigFile = awbConfigFile;
    }

    private File awbConfigFile = new File("");

    @Config(message = "Automatically generate the bundle's packageId", order = 6, advance = false, group = "atlas")
    private boolean autoPackageId = true;

    @Config(message = "Automatic allocation of packageId The minimum value", order = 6, advance = false, group = "atlas")
    private int minPackageId = 35;

    @Config(title = "Build base pack", message = "Build the baseline package, and the suggestion is open, otherwise the patch package will not be able to proceed", order = 0, group = "atlas_patch")
    private Boolean createAP = true;

    @Config(message = "Merge the bundle jarResource file", order = 8, advance = true, group = "atlas")
    private Boolean mergeAwbJavaRes = false;

    @Config(message = "Whether to rely on conflict to terminate packaging", order = 0, group = "check")
    private boolean abortIfDependencyConflict = false;

    @Config(message = "Whether class conflicts terminate packaging", order = 0, group = "check")
    private boolean abortIfClassConflict = false;

    @Config(title = "Pretreatment of the manifest", message = "If atlas is turned on, it must be true", order = 7, advance = true, group = "atlas")
    private Boolean preProcessManifest = true;

    @Config(title = "Use a custom aapt", message = "If atlas is turned on, it must be true", order = 8, advance = true, group = "atlas")
    private Boolean useCustomAapt = true;

    @Config(title = "aaptThe output R is constant", message = "The recommended value is set to false, You can reduce the amount of patch packages that are dynamically deployed", order = 9, advance = true,
        group = "atlas")
    private Boolean aaptConstantId = true;

    @Config(message = "Inject the core bundle information", advance = true, order = 10, group = "atlas")
    private Boolean classInject = false;

    @Config(title = "The main dex in pile", message = "The old version of the main dex dynamic deployment, has been deprecated", advance = true, order = 11, group = "atlas")
    private Boolean doPreverify = false;

    @Deprecated
    private Boolean resV4Enabled = true;


    public Boolean getDisabledBundleDependency() {
        return disabledBundleDependency;
    }

    public void setDisabledBundleDependency(Boolean disabledBundleDependency) {
        this.disabledBundleDependency = disabledBundleDependency;
    }

    private Boolean disabledBundleDependency = false;

    @Config(message = "classInjection before proguard", advance = true, order = 12, group = "atlas")
    private Boolean injectBeforeProguard = false;

    @Config(title = "Use the databinding bundle list", message = "You need to do a databinding bundle, A value of packageName ", order = 13,
        advance = true, group = "atlas")
    private Set<String> dataBindingBundles = new HashSet<>();

    @Config(message = "proguardDo you need to read the confusing configuration in the bundle", order = 14, advance = true, group = "atlas")
    private boolean bundleProguardConfigEnabled = true;

    @Config(message = "The confusion in dependencies reads only the keep rule", order = 15, advance = true, group = "atlas")
    private boolean libraryProguardKeepOnly = true;

    @Config(title = "proguardConfigure read dependency blacklist", message = "group:name,group2:name2", order = 16, advance = true,
        group = "atlas")
    private Set<String> bundleProguardConfigBlackList = new HashSet<>();

    @Deprecated
    private Set<String> insideOfApkBundles = Sets.newHashSet();


    public Set<String> getInjectExcludePkgs() {
        return injectExcludePkgs;
    }

    public void setInjectExcludePkgs(Set<String> injectExcludePkgs) {
        this.injectExcludePkgs = injectExcludePkgs;
    }

    private Set<String>injectExcludePkgs = new HashSet<>();


    public Set<String> getInjectPkgs() {
        return injectPkgs;
    }

    public void setInjectPkgs(Set<String> injectPkgs) {
        this.injectPkgs = injectPkgs;
    }

    private Set<String>injectPkgs = new HashSet<>();


    public Set<String> getOutOfApkNativeSos() {
        return outOfApkNativeSos;
    }

    public void setOutOfApkNativeSos(Set<String> outOfApkNativeSos) {
        this.outOfApkNativeSos = outOfApkNativeSos;
    }

    private Set<String>outOfApkNativeSos = Sets.newHashSet();

    private boolean incremental = false;


    public boolean isInjectSerialVersionUID() {
        return injectSerialVersionUID;
    }

    public void setInjectSerialVersionUID(boolean injectSerialVersionUID) {
        this.injectSerialVersionUID = injectSerialVersionUID;
    }

    private boolean injectSerialVersionUID = false;

    @Config(message = "Whether to use fast proguard", order = 16, advance = true, group = "atlas")
    private boolean fastProguard = false;

    private boolean keepJavaResAfterProguard = false;

    @Config(message = "Whether to use the proguard cache", order = 17, advance = true, group = "atlas")
    private boolean proguardCacheEnabled = true;

    @Config(message = "Whether to use the remote proguard cache", order = 16, advance = true, group = "atlas")
    private boolean proguardNetworkCacheEnabled = false;

    @Config(message = "Whether to use the dex cache", order = 17, advance = true, group = "atlas")
    private boolean dexCacheEnabled = true;

    @Config(message = "Whether to use the remote dex cache", order = 18, advance = true, group = "atlas")
    private boolean dexNetworkCacheEnabled = false;

    private int proguardParallelCount = 8;

    @Config(message = "put awb so in assets, not  lib/armeabi dir", order = 19, advance = true, group = "atlas")
    private Set<String> keepInAssetsSoNames = new HashSet<>();

    @Config(message = "put awb so in lib/armeabi , not assets dir", order = 19, advance = true, group = "atlas")
    private Set<String> keepInLibSoNames = new HashSet<>();


    public Set<String> getRemoveSoFiles() {
        return removeSoFiles;
    }

    public void setRemoveSoFiles(Set<String> removeSoFiles) {
        this.removeSoFiles = removeSoFiles;
    }

    public File getPackageIdFile() {
        return packageIdFile;
    }

    public void setPackageIdFile(File packageIdFile) {
        this.packageIdFile = packageIdFile;
    }

    public boolean isAutoPackageId() {
        return autoPackageId;
    }

    public void setAutoPackageId(boolean autoPackageId) {
        this.autoPackageId = autoPackageId;
    }

    public Boolean getPreProcessManifest() {
        return preProcessManifest;
    }

    public void setPreProcessManifest(Boolean preProcessManifest) {
        this.preProcessManifest = preProcessManifest;
    }

    public Boolean getUseCustomAapt() {
        return useCustomAapt;
    }

    public void setUseCustomAapt(Boolean useCustomAapt) {
        this.useCustomAapt = useCustomAapt;
    }

    public Boolean getAaptConstantId() {
        return aaptConstantId;
    }

    public void setAaptConstantId(Boolean aaptConstantId) {
        this.aaptConstantId = aaptConstantId;
    }

    public Boolean getClassInject() {
        return classInject;
    }

    public void setClassInject(Boolean classInject) {
        this.classInject = classInject;
    }

    public Boolean getInjectBeforeProguard() {
        return injectBeforeProguard;
    }

    public void setInjectBeforeProguard(Boolean injectBeforeProguard) {
        this.injectBeforeProguard = injectBeforeProguard;
    }

    public Boolean getCreateAP() {
        return createAP;
    }

    public Boolean getMergeAwbJavaRes() {
        return mergeAwbJavaRes;
    }

    public void setMergeAwbJavaRes(Boolean mergeAwbJavaRes) {
        this.mergeAwbJavaRes = mergeAwbJavaRes;
    }

    public Set<String> getOutOfApkBundles() {

        return outOfApkBundles;
    }

    public void setOutOfApkBundles(Set<String> outOfApkBundles) {
        this.outOfApkBundles = outOfApkBundles;
    }

    public Set<String> getInsideOfApkBundles() {
        return insideOfApkBundles;
    }

    public void setInsideOfApkBundles(Set<String> insideOfApkBundles) {
        this.insideOfApkBundles = insideOfApkBundles;
    }

    public List<String> getAutoStartBundles() {
        return autoStartBundles;
    }

    public void setAutoStartBundles(List<String> autoStartBundles) {
        this.autoStartBundles = autoStartBundles;
    }

    public String getPreLaunch() {
        return preLaunch;
    }

    public void setPreLaunch(String preLaunch) {
        this.preLaunch = preLaunch;
    }

    public Boolean getDoPreverify() {
        return doPreverify;
    }

    public void setDoPreverify(Boolean doPreverify) {
        this.doPreverify = doPreverify;
    }

    public Boolean getResV4Enabled() {
        return resV4Enabled;
    }

    public void setResV4Enabled(Boolean resV4Enabled) {
        this.resV4Enabled = resV4Enabled;
    }

    public boolean isAbortIfDependencyConflict() {
        return abortIfDependencyConflict;
    }

    public void setAbortIfDependencyConflict(boolean abortIfDependencyConflict) {
        this.abortIfDependencyConflict = abortIfDependencyConflict;
    }

    public boolean isAbortIfClassConflict() {
        return abortIfClassConflict;
    }

    public void setAbortIfClassConflict(boolean abortIfClassConflict) {
        this.abortIfClassConflict = abortIfClassConflict;
    }

    public Set<String> getDataBindingBundles() {
        return dataBindingBundles;
    }

    public void setDataBindingBundles(Set<String> dataBindingBundles) {
        this.dataBindingBundles = dataBindingBundles;
    }

    public boolean isAtlasMultiDex() {
        return atlasMultiDex;
    }

    public void setAtlasMultiDex(boolean atlasMultiDex) {
        this.atlasMultiDex = atlasMultiDex;
    }

    public Boolean isCreateAP() {
        return createAP;
    }

    public void setCreateAP(Boolean createAP) {
        this.createAP = createAP;
    }

    public boolean isIncremental() {
        return incremental;
    }

    public void setIncremental(boolean incremental) {
        this.incremental = incremental;
    }

    public boolean isBundleProguardConfigEnabled() {
        return bundleProguardConfigEnabled;
    }

    public void setBundleProguardConfigEnabled(boolean bundleProguardConfigEnabled) {
        this.bundleProguardConfigEnabled = bundleProguardConfigEnabled;
    }

    public boolean isLibraryProguardKeepOnly() {
        return libraryProguardKeepOnly;
    }

    public void setLibraryProguardKeepOnly(boolean libraryProguardKeepOnly) {
        this.libraryProguardKeepOnly = libraryProguardKeepOnly;
    }

    public Set<String> getBundleProguardConfigBlackList() {
        return bundleProguardConfigBlackList;
    }

    public void setBundleProguardConfigBlackList(Set<String> bundleProguardConfigBlackList) {
        this.bundleProguardConfigBlackList = bundleProguardConfigBlackList;
    }

    public int getMinPackageId() {
        return minPackageId;
    }

    public void setMinPackageId(int minPackageId) {
        this.minPackageId = minPackageId;
    }

    public boolean isFastProguard() {
        return fastProguard;
    }

    public void setFastProguard(boolean fastProguard) {
        this.fastProguard = fastProguard;
    }

    public int getProguardParallelCount() {
        return proguardParallelCount;
    }

    public void setProguardParallelCount(int proguardParallelCount) {
        this.proguardParallelCount = proguardParallelCount;
    }

    public boolean isProguardCacheEnabled() {
        return proguardCacheEnabled;
    }

    public void setProguardCacheEnabled(boolean proguardCacheEnabled) {
        this.proguardCacheEnabled = proguardCacheEnabled;
    }

    public boolean isProguardNetworkCacheEnabled() {
        return proguardNetworkCacheEnabled;
    }

    public void setProguardNetworkCacheEnabled(boolean proguardNetworkCacheEnabled) {
        this.proguardNetworkCacheEnabled = proguardNetworkCacheEnabled;
    }

    public boolean isDexCacheEnabled() {
        return dexCacheEnabled;
    }

    public void setDexCacheEnabled(boolean dexCacheEnabled) {
        this.dexCacheEnabled = dexCacheEnabled;
    }

    public boolean isDexNetworkCacheEnabled() {
        return dexNetworkCacheEnabled;
    }

    public void setDexNetworkCacheEnabled(boolean dexNetworkCacheEnabled) {
        this.dexNetworkCacheEnabled = dexNetworkCacheEnabled;
    }

    public Set<String> getKeepInAssetsSoNames() {
        return keepInAssetsSoNames;
    }

    public void setKeepInAssetsSoNames(Set<String> keepInAssetsSoNames) {
        this.keepInAssetsSoNames = keepInAssetsSoNames;
    }

    public Set<String> getKeepInLibSoNames() {
        return keepInLibSoNames;
    }

    public void setKeepInLibSoNames(Set<String> keepInLibSoNames) {
        this.keepInLibSoNames = keepInLibSoNames;
    }

    public boolean isKeepJavaResAfterProguard() {
        return keepJavaResAfterProguard;
    }

    public void setKeepJavaResAfterProguard(boolean keepJavaResAfterProguard) {
        this.keepJavaResAfterProguard = keepJavaResAfterProguard;
    }

    public void setPushInstall(boolean pushInstall) {
        this.pushInstall = pushInstall;
    }

    private boolean pushInstall;
    public boolean isPushInstall() {
        return pushInstall;
    }

    public boolean isPatchConstructors() {
        return patchConstructors;
    }

    public void setPatchConstructors(boolean patchConstructors) {
        this.patchConstructors = patchConstructors;
    }

    private boolean patchConstructors = false;

    public boolean isPatchEachMethod() {
        return patchEachMethod;
    }

    public void setPatchEachMethod(boolean patchEachMethod) {
        this.patchEachMethod = patchEachMethod;
    }

    private boolean patchEachMethod = false;



    public int getPatchSuperMethodCount() {
        return patchSuperMethodCount;
    }

    public void setPatchSuperMethodCount(int patchSuperMethodCount) {
        this.patchSuperMethodCount = patchSuperMethodCount;
    }

    private int patchSuperMethodCount = 100;

    public void setSupportAddCallSuper(boolean supportAddCallSuper) {
        this.supportAddCallSuper = supportAddCallSuper;
    }

    private boolean supportAddCallSuper = false;



    public boolean isSupportAddCallSuper() {
        return supportAddCallSuper;
    }
}
