/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.manager;

import com.android.SdkConstants;
import com.android.build.api.transform.QualifiedContent;
import com.android.build.gradle.AppExtension;
import com.android.build.gradle.api.ApplicationVariant;
import com.android.build.gradle.api.BaseVariantOutput;
import com.android.build.gradle.internal.ApkDataUtils;
import com.android.build.gradle.internal.TaskContainerAdaptor;
import com.android.build.gradle.internal.TaskFactory;
import com.android.build.gradle.internal.api.AppVariantContext;
import com.android.build.gradle.internal.api.ApplicationVariantImpl;
import com.android.build.gradle.internal.api.BaseVariantOutputImpl;
import com.android.build.gradle.internal.core.GradleVariantConfiguration;
import com.android.build.gradle.internal.incremental.InstantRunPatchingPolicy;
import com.android.build.gradle.internal.pipeline.OriginalStream;
import com.android.build.gradle.internal.pipeline.TransformTask;
import com.android.build.gradle.internal.scope.GlobalScope;
import com.android.build.gradle.internal.scope.VariantScope;
import com.android.build.gradle.internal.tasks.ExtractProguardFiles;
import com.android.build.gradle.internal.tasks.ExtractTryWithResourcesSupportJar;
import com.android.build.gradle.internal.transforms.*;
import com.android.build.gradle.options.BooleanOption;
import com.android.build.gradle.options.ProjectOptions;
import com.android.build.gradle.tasks.*;
import com.android.builder.core.AtlasBuilder;
import com.android.builder.core.DefaultDexOptions;
import com.android.builder.core.DexByteCodeConverter;
import com.android.builder.core.ErrorReporter;
import com.android.builder.dexing.DexingType;
import com.android.builder.sdk.TargetInfo;
import com.android.ide.common.build.ApkData;
import com.google.common.base.Supplier;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Sets;
import com.taobao.android.builder.AtlasBuildContext;
import com.taobao.android.builder.AtlasMainDexHelper;
import com.taobao.android.builder.dependency.AtlasDependencyTree;
import com.taobao.android.builder.dependency.model.AwbBundle;
import com.taobao.android.builder.extension.AtlasExtension;
import com.taobao.android.builder.insant.DelegateProguardTransform;
import com.taobao.android.builder.tasks.PrepareAPTask;
import com.taobao.android.builder.tasks.app.*;
import com.taobao.android.builder.tasks.app.BuildAtlasEnvTask;
import com.taobao.android.builder.tasks.app.bundle.JavacAwbsTask;
import com.taobao.android.builder.tasks.app.bundle.PackageAwbsTask;
import com.taobao.android.builder.tasks.app.bundle.ProcessResAwbsTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingExportBuildInfoTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingMergeArtifactsTask;
import com.taobao.android.builder.tasks.app.databinding.AwbDataBindingRenameTask;
import com.taobao.android.builder.tasks.app.manifest.StandardizeLibManifestTask;
import com.taobao.android.builder.tasks.app.merge.*;
import com.taobao.android.builder.tasks.app.merge.bundle.MergeAwbAssets;
import com.taobao.android.builder.tasks.app.prepare.PrepareAaptTask;
import com.taobao.android.builder.tasks.app.prepare.PrepareBundleInfoTask;
import com.taobao.android.builder.tasks.app.prepare.PreparePackageIdsTask;
import com.taobao.android.builder.tasks.instantapp.AtlasBundleInstantApp;
import com.taobao.android.builder.tasks.manager.MtlTaskContext;
import com.taobao.android.builder.tasks.manager.MtlTaskInjector;
import com.taobao.android.builder.tasks.manager.transform.MtlTransformContext;
import com.taobao.android.builder.tasks.manager.transform.MtlTransformInjector;
import com.taobao.android.builder.tasks.manager.transform.TransformManager;
import com.taobao.android.builder.tasks.tpatch.DiffBundleInfoTask;
import com.taobao.android.builder.tasks.tpatch.TPatchDiffApkBuildTask;
import com.taobao.android.builder.tasks.tpatch.TPatchDiffResAPBuildTask;
import com.taobao.android.builder.tasks.tpatch.TPatchTask;
import com.taobao.android.builder.tasks.transform.AtlasDesugarTransform;
import com.taobao.android.builder.tasks.transform.AtlasProguardTransform;
import com.taobao.android.builder.tasks.transform.ClassInjectTransform;
import com.taobao.android.builder.tasks.transform.TransformReplacer;
import com.taobao.android.builder.tools.ReflectUtils;
import com.taobao.android.builder.tools.multidex.FastMultiDexer;
import groovy.lang.Closure;
import org.apache.commons.io.FileUtils;
import org.gradle.api.Action;
import org.gradle.api.GradleException;
import org.gradle.api.Project;
import org.gradle.api.Task;
import org.gradle.api.file.ConfigurableFileCollection;
import org.gradle.api.file.FileCollection;
import org.gradle.api.file.FileTree;
import org.gradle.api.internal.AbstractTask;
import org.gradle.api.internal.file.AbstractFileCollection;
import org.gradle.api.internal.tasks.TaskExecuter;
import org.gradle.api.internal.tasks.execution.ExecuteActionsTaskExecuter;
import org.gradle.api.specs.Spec;
import org.gradle.api.tasks.StopExecutionException;
import org.gradle.api.tasks.TaskCollection;
import org.gradle.api.tasks.TaskDependency;
import org.gradle.api.tasks.compile.JavaCompile;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.lang.reflect.Field;
import java.util.*;
import java.util.function.Consumer;

import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ArtifactScope.ALL;
import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ArtifactType.JAR;
import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ConsumedConfigType.ANNOTATION_PROCESSOR;

/**
 * MTLThe plug-in compiles apk's task management
 * Created by shenghua.nish on 2016-05-09 3:55 in the afternoon.
 */
@SuppressWarnings("ALL")
public class AtlasAppTaskManager extends AtlasBaseTaskManager {

    private AppExtension appExtension;

    public AtlasAppTaskManager(AtlasBuilder androidBuilder, AppExtension appExtension, Project project,
                               AtlasExtension atlasExtension) {
        super(androidBuilder, appExtension, project, atlasExtension);
        this.appExtension = appExtension;
    }

    @Override
    public void runTask() {

        appExtension.getApplicationVariants().forEach(new Consumer<ApplicationVariant>() {

                                                          @Override
                                                          public void accept(ApplicationVariant applicationVariant) {


                                                              AppVariantContext appVariantContext = AtlasBuildContext.sBuilderAdapter.appVariantContextFactory
                                                                      .getAppVariantContext(project, applicationVariant);
                                                              if (!AtlasBuildContext.atlasMainDexHelperMap.containsKey(appVariantContext.getVariantName())){
                                                                  AtlasBuildContext.atlasMainDexHelperMap.put(appVariantContext.getVariantName(),new AtlasMainDexHelper());
                                                              }

                                                              TransformReplacer transformReplacer = new TransformReplacer(appVariantContext);

                                                              repalceAndroidBuilder(applicationVariant);

                                                              List<MtlTaskContext> mtlTaskContextList = new ArrayList<MtlTaskContext>();

                                                              mtlTaskContextList.add(new MtlTaskContext(appVariantContext.getVariantData().preBuildTask));

                                                              mtlTaskContextList.add(new MtlTaskContext(BuildAtlasEnvTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(ScanDupResTask.ConfigActon.class,null));

                                                              mtlTaskContextList.add(new MtlTaskContext(LogDependenciesTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(PrepareAPTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(appVariantContext.getVariantData().mergeAssetsTask));

                                                              mtlTaskContextList.add(new MtlTaskContext(RenderscriptCompile.class));

                                                              mtlTaskContextList.add(new MtlTaskContext(StandardizeLibManifestTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(PrepareBundleInfoTask.ConfigAction.class, null));

                                                              if (!atlasExtension.getTBuildConfig().getClassInject() && atlasExtension.isAtlasEnabled()) {
                                                                  mtlTaskContextList.add(new MtlTaskContext(GenerateAtlasSourceTask.ConfigAction.class, null));
                                                              }

                                                              mtlTaskContextList.add(new MtlTaskContext(PreparePackageIdsTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(PrepareAaptTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(AidlCompile.class));

                                                              mtlTaskContextList.add(new MtlTaskContext(GenerateBuildConfig.class));

                                                              mtlTaskContextList.add(new MtlTaskContext(MergeResAwbsConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(MergeAssetAwbsConfigAction.class, null));

                                                              if (null != androidExtension.getDataBinding() && androidExtension.getDataBinding().isEnabled() && !appVariantContext.getAtlasExtension().getTBuildConfig().getAllBundlesToMdex()) {

//                                                                  mtlTaskContextList.add(
//                                                                          new MtlTaskContext(AwbDataBindingProcessLayoutTask.ConfigAction.class, null));
                                                                  mtlTaskContextList.add(
                                                                          new MtlTaskContext(AwbDataBindingExportBuildInfoTask.ConfigAction.class, null));
                                                                  mtlTaskContextList.add(
                                                                          new MtlTaskContext(AwbDataBindingMergeArtifactsTask.ConfigAction.class, null));


                                                              }

                                                              mtlTaskContextList.add(new MtlTaskContext(MergeManifests.class));

                                                              mtlTaskContextList.add(new MtlTaskContext(MergeManifestAwbsConfigAction.class, null));

                                                              //mtlTaskContextList.add(new MtlTaskContext(MergeResV4Dir.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(ProcessAndroidResources.class));

                                                              ProcessAndroidResources processAndroidResources
                                                                      = appVariantContext.getScope()
                                                                      .getProcessResourcesTask()
                                                                      .get(new TaskContainerAdaptor(appVariantContext.getProject()
                                                                              .getTasks()));
                                                              if (processAndroidResources.isAapt2Enabled()) {
                                                                  processAndroidResources.doLast(new Action<Task>() {
                                                                      @Override
                                                                      public void execute(Task task) {
                                                                          File processResourcePackageOutputDirectory
                                                                                  = appVariantContext.getScope()
                                                                                  .getProcessResourcePackageOutputDirectory();
                                                                          File[] files
                                                                                  = processResourcePackageOutputDirectory.listFiles(
                                                                                  (file, name) -> name.endsWith(SdkConstants.DOT_RES));
                                                                          for (File file : files) {
                                                                              try {
                                                                                  ResourcePatch.makePatchable(file);
                                                                              } catch (IOException e) {
                                                                                  throw new UncheckedIOException(e);
                                                                              }
                                                                          }
                                                                      }
                                                                  });
                                                              }
                                                              mtlTaskContextList.add(new MtlTaskContext(ProcessResAwbsTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(JavacAwbsTask.ConfigAction.class, null));


                                                              if (null != androidExtension.getDataBinding() && androidExtension.getDataBinding().isEnabled() && !appVariantContext.getAtlasExtension().getTBuildConfig().getAllBundlesToMdex()) {
                                                                  mtlTaskContextList.add(new MtlTaskContext(AwbDataBindingRenameTask.ConfigAction.class, null));
                                                              }

                                                              mtlTaskContextList.add(new MtlTaskContext(TransformTask.class));

                                                              mtlTaskContextList.add(new MtlTaskContext(PackageAwbsTask.ConfigAction.class, null));

                                                              if (appVariantContext.getAtlasExtension().getTBuildConfig().isIncremental() && (
                                                                      appVariantContext.getBuildType().getPatchConfig() == null || !appVariantContext.getBuildType()
                                                                              .getPatchConfig().isCreateTPatch())) {
//                                                                  mtlTaskContextList.add(new MtlTaskContext(PrepareBaseApkTask.ConfigAction.class, null));
                                                                  final TaskFactory tasks = new TaskContainerAdaptor(project.getTasks());
                                                                  VariantScope variantScope = appVariantContext.getVariantData().getScope();


                                                                  // create the stream generated from this task
                                                                  variantScope.getTransformManager().addStream(OriginalStream.builder(project, applicationVariant.getName()).addContentType(
                                                                          QualifiedContent.DefaultContentType.RESOURCES).addScope(QualifiedContent.Scope.PROJECT)
                                                                          .setFolders(new Supplier<Collection<File>>() {
                                                                              @Override
                                                                              public Collection<File> get() {
                                                                                  return ImmutableList.of(new File(
                                                                                          appVariantContext.apContext
                                                                                                  .getBaseApk() + "_"));
                                                                              }
                                                                          })
                                                                          .build());
                                                              }


                                                              final TaskFactory tasks = new TaskContainerAdaptor(project.getTasks());
                                                              VariantScope variantScope = appVariantContext.getVariantData().getScope();

                                                              mtlTaskContextList.add(new MtlTaskContext(PackageApplication.class));

                                                              if (appVariantContext.getAtlasExtension().isInstantAppEnabled()) {

                                                                  mtlTaskContextList.add(new MtlTaskContext(AtlasBundleInstantApp.ConfigAction.class, null));
                                                              }

                                                              mtlTaskContextList.add(new MtlTaskContext(ApBuildTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(DiffBundleInfoTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(TPatchDiffResAPBuildTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(TPatchDiffApkBuildTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext(TPatchTask.ConfigAction.class, null));

                                                              mtlTaskContextList.add(new MtlTaskContext("assemble"));

                                                              new MtlTaskInjector(appVariantContext).injectTasks(mtlTaskContextList, tAndroidBuilder);

                                                              List<MtlTransformContext> mtlTransformContextList = new ArrayList<MtlTransformContext>();

                                                              if (atlasExtension.getTBuildConfig().getClassInject()) {
                                                                  mtlTransformContextList.add(
                                                                          new MtlTransformContext(ClassInjectTransform.class, ProGuardTransform.class,
                                                                                  DexTransform.class));
                                                              }
                                                              if (variantScope.getInstantRunBuildContext().isInInstantRunMode() && appVariantContext.getVariantConfiguration().getBuildType().isMinifyEnabled()){
                                                                  mtlTransformContextList.add(
                                                                          new MtlTransformContext(DelegateProguardTransform.class, ExtractJarsTransform.class,
                                                                                  InstantRunTransform.class));
                                                              }


                                                              if (!mtlTransformContextList.isEmpty()) {
                                                                  new MtlTransformInjector(appVariantContext).injectTasks(mtlTransformContextList);
                                                              }
                                                              Collection<BaseVariantOutput> baseVariantOutputDataList = appVariantContext.getVariantOutputData();

                                                              boolean multiDexEnabled = appVariantContext.getVariantData().getVariantConfiguration().isMultiDexEnabled();

                                                              if (atlasExtension.getTBuildConfig().isAtlasMultiDex() && multiDexEnabled) {
                                                                  transformReplacer.replaceMultiDexListTransform();
                                                              }


                                                              transformReplacer.replaceProguardTransform();

                                                              transformReplacer.disableCache();


                                                              if (variantScope.getGlobalScope().getExtension().getDataBinding().isEnabled() && !appVariantContext.getAtlasExtension().getTBuildConfig().getAllBundlesToMdex()) {
                                                                  transformReplacer.replaceDataBindingMergeArtifactsTransform();
                                                              }

                                                              for (final BaseVariantOutput vod : baseVariantOutputDataList) {
                                                                  transformReplacer.replaceFixStackFramesTransform(vod);
                                                                  transformReplacer.replaceDesugarTransform(vod);
                                                                  transformReplacer.replaceDexArchiveBuilderTransform(vod);
                                                                  transformReplacer.replaceDexExternalLibMerge(vod);
                                                                  transformReplacer.replaceDexMerge(vod);
                                                                  transformReplacer.replaceDexTransform(appVariantContext, vod);
                                                                  transformReplacer.replaceShrinkResourcesTransform();
                                                                  transformReplacer.replaceMergeJavaResourcesTransform(appVariantContext, vod);
                                                                  transformReplacer.repalaceSomeInstantTransform(vod);

                                                                  if (atlasExtension.getTBuildConfig().isIncremental()) {
                                                                      InstantRunPatchingPolicy patchingPolicy = variantScope.getInstantRunBuildContext()
                                                                              .getPatchingPolicy();
                                                                      BaseVariantOutputImpl variantOutput = (BaseVariantOutputImpl) vod;
                                                                      ApkData data = ApkDataUtils.get(variantOutput);
                                                                      AtlasDependencyTree atlasDependencyTree = AtlasBuildContext.androidDependencyTrees.get(
                                                                              variantScope.getFullVariantName());

//
                                                                  }

                                                              }



//                                                              Boolean includeCompileClasspath =
//                                                                      appVariantContext.getScope().getVariantConfiguration()
//                                                                              .getJavaCompileOptions()
//                                                                              .getAnnotationProcessorOptions()
//                                                                              .getIncludeCompileClasspath();

//                                                              appVariantContext.getVariantData().javaCompilerTask.doFirst(task -> {
//                                                                  JavaCompile compile = (JavaCompile) task;
//                                                                  Set<File> mainDexFiles = new MainFilesCollection(appVariantContext.getVariantName()).getFiles();
//                                                                  FileCollection mainFiles = appVariantContext.getProject().files(mainDexFiles);
//                                                                  FileCollection files = appVariantContext.getScope().getArtifactFileCollection(ANNOTATION_PROCESSOR, ALL, JAR);
//                                                                  FileCollection bootFiles = appVariantContext.getProject().files(appVariantContext.getScope().getGlobalScope().getAndroidBuilder().getBootClasspath(false));
//                                                                  mainFiles = mainFiles.plus(bootFiles);
//                                                                  FileCollection fileCollection = compile.getClasspath();
//                                                                  File kotlinClasses = null;
//                                                                  for (File file : fileCollection) {
//                                                                      if (file.getAbsolutePath().contains("kotlin-classes")) {
//                                                                          mainFiles = mainFiles.plus(appVariantContext.getProject().files(file));
//                                                                          kotlinClasses = file;
//                                                                          break;
//                                                                      }
//                                                                  }
//                                                                  compile.setClasspath(mainFiles);
//                                                                  if (Boolean.TRUE.equals(includeCompileClasspath)) {
//                                                                      compile.getOptions().setAnnotationProcessorPath(files.plus(mainFiles));
//                                                                  }
//                                                              });

                                                              appVariantContext.getVariantData().javaCompilerTask.doLast(new Action<Task>() {
                                                                  @Override
                                                                  public void execute(Task task) {
                                                                      JavaCompile compile = (JavaCompile) task;
                                                                      AtlasBuildContext.atlasMainDexHelperMap.get(appVariantContext.getVariantName()).getInputDirs().add(compile.getDestinationDir());
                                                                  }
                                                              });

                                                              PackageAndroidArtifact packageAndroidArtifact = appVariantContext.getVariantData().getTaskByType(PackageAndroidArtifact.class);
                                                              if (packageAndroidArtifact != null) {
                                                                  ReflectUtils.updateField(packageAndroidArtifact, "javaResourceFiles", new AbstractFileCollection() {
                                                                      @Override
                                                                      public String getDisplayName() {
                                                                          return "java-merge-res.jar";
                                                                      }

                                                                      @Override
                                                                      public Set<File> getFiles() {
                                                                          if (AtlasBuildContext.atlasMainDexHelperMap.get(packageAndroidArtifact.getVariantName()).getMainJavaRes() == null) {
                                                                              return Sets.newHashSet();
                                                                          }
                                                                          return Sets.newHashSet(AtlasBuildContext.atlasMainDexHelperMap.get(packageAndroidArtifact.getVariantName()).getMainJavaRes());
                                                                      }
                                                                  });
                                                              }

                                                              TaskCollection<ExtractTryWithResourcesSupportJar>taskCollection = appVariantContext.getProject().getTasks().withType(ExtractTryWithResourcesSupportJar.class);
                                                              for (ExtractTryWithResourcesSupportJar task : taskCollection) {
                                                                  task.doLast(new Action<Task>() {
                                                                      @Override
                                                                      public void execute(Task task) {
                                                                          ConfigurableFileCollection fileCollection = variantScope.getTryWithResourceRuntimeSupportJar();
                                                                          for (File file : fileCollection.getFiles()) {
                                                                              if (file.exists()) {
                                                                                  AtlasBuildContext.atlasMainDexHelperMap.get(variantScope.getFullVariantName()).addMainDex(new BuildAtlasEnvTask.FileIdentity("runtime-deps-try-with-resources", file, false, false));
                                                                                  break;
                                                                              }
                                                                          }
                                                                      }
                                                                  });

                                                              }
                                                          }
                                                      }
        );
    }






    private void repalceAndroidBuilder(ApplicationVariant applicationVariant) {
        try {
            if (applicationVariant instanceof ApplicationVariantImpl) {
                VariantScope variantScope = ((ApplicationVariantImpl) applicationVariant).getVariantData().getScope();
                GlobalScope globalScope = variantScope.getGlobalScope();
//                ReflectUtils.updateField(globalScope, "androidBuilder", AtlasBuildContext.androidBuilderMap.get(globalScope.getProject()));
//                Field f = ProjectOptions.class.getDeclaredField("booleanOptions");
//                f.setAccessible(true);
//                Map map = (Map) f.get(((ApplicationVariantImpl) applicationVariant).getVariantData().getScope().getGlobalScope().getProjectOptions());
//                map.put(BooleanOption.ENABLE_AAPT2, false);
                AtlasBuildContext.androidBuilderMap.get(globalScope.getProject()).initAapt(globalScope.getProjectOptions());

            }
        } catch (Exception e) {
            logger.error(e, "write globalScope androidBuilder field failed!");
        }
    }


}
