/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.tasks.app.bundle;

import com.android.build.api.transform.QualifiedContent;
import com.android.build.gradle.api.AnnotationProcessorOptions;
import com.android.build.gradle.internal.CompileOptions;
import com.android.build.gradle.internal.LoggerWrapper;
import com.android.build.gradle.internal.api.AppVariantOutputContext;
import com.android.build.gradle.internal.pipeline.OriginalStream;
import com.android.build.gradle.internal.publishing.AndroidArtifacts;
import com.android.build.gradle.internal.scope.ConventionMappingHelper;
import com.android.build.gradle.internal.scope.TaskConfigAction;
import com.android.build.gradle.internal.scope.VariantScope;
import com.android.build.gradle.tasks.factory.AbstractCompilesUtil;
import com.android.build.gradle.tasks.factory.AndroidJavaCompile;
import com.android.build.gradle.tasks.factory.AwbAndroidJavaCompile;
import com.android.builder.profile.ProcessProfileWriter;
import com.android.utils.FileUtils;
import com.android.utils.ILogger;
import com.google.common.base.Joiner;
import com.google.wireless.android.sdk.stats.GradleBuildProject;
import com.google.wireless.android.sdk.stats.GradleBuildVariant;
import com.taobao.android.builder.dependency.model.AwbBundle;
import com.taobao.android.builder.tools.ReflectUtils;
import org.gradle.api.Action;
import org.gradle.api.Task;
import org.gradle.api.artifacts.Configuration;
import org.gradle.api.file.FileCollection;
import org.gradle.api.tasks.compile.JavaCompile;

import java.io.File;
import java.io.IOException;
import java.sql.Ref;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;

import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ArtifactScope.ALL;
import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ArtifactType.*;
import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ConsumedConfigType.ANNOTATION_PROCESSOR;
import static com.android.build.gradle.internal.publishing.AndroidArtifacts.ConsumedConfigType.COMPILE_CLASSPATH;
import static com.android.build.gradle.internal.scope.TaskOutputHolder.TaskOutputType.ANNOTATION_PROCESSOR_LIST;
import static com.android.build.gradle.internal.scope.TaskOutputHolder.TaskOutputType.DATA_BINDING_DEPENDENCY_ARTIFACTS;
/**
 * @author lilong
 * @create 2017-12-08 上午3:43
 */

public class AwbJavaCompileConfigAction implements TaskConfigAction<AwbAndroidJavaCompile> {

    private static final ILogger LOG = LoggerWrapper.getLogger(AwbJavaCompileConfigAction.class);

    private AwbBundle awbBundle;

    private AppVariantOutputContext appVariantOutputContext;

    private VariantScope scope;

    public AwbJavaCompileConfigAction(AwbBundle awbBundle,
                                      AppVariantOutputContext appVariantOutputContext) {
        this.awbBundle = awbBundle;
        this.appVariantOutputContext = appVariantOutputContext;
        this.scope = appVariantOutputContext.getScope();
    }

    @Override
    public String getName() {
        return scope.getTaskName("compileAwb", "JavaWithJavac[" + awbBundle.getName() + "]");
    }

    @Override
    public Class<AwbAndroidJavaCompile> getType() {
        return AwbAndroidJavaCompile.class;
    }

    /**
     * AwbInstantRun is not supported for the moment
     *
     * @param javacTask
     */
    @Override
    public void execute(AwbAndroidJavaCompile javacTask) {

        appVariantOutputContext.getAwbJavacTasks().put(awbBundle.getName(), javacTask);
        ProcessAwbAndroidResources processAwbAndroidResources = appVariantOutputContext.getAwbAndroidResourcesMap()
                .get(awbBundle.getName());
        assert null != processAwbAndroidResources;

        javacTask.source(processAwbAndroidResources.getSourceOutputDir());
        if (appVariantOutputContext.getVariantContext().isDataBindEnabled(awbBundle) && awbBundle.isDataBindEnabled()) {
            javacTask.source(appVariantOutputContext.getVariantContext()
                    .getAwbClassOutputForDataBinding(awbBundle));
        }


//        ConventionMappingHelper.map(javacTask, "classpath", (Callable<FileCollection>) () -> getInputJars());

        FileCollection classpath = getInputJars();

        javacTask.setDestinationDir(appVariantOutputContext.getJAwbavaOutputDir(awbBundle));

        final boolean keepDefaultBootstrap = scope.keepDefaultBootstrap();

        if (!keepDefaultBootstrap) {
            // Set boot classpath if we don't need to keep the default.  Otherwise, this is added as
            // normal classpath.
            javacTask
                    .getOptions()
                    .setBootClasspath(
                            Joiner.on(File.pathSeparator)
                                    .join(
                                            scope.getGlobalScope()
                                                    .getAndroidBuilder()
                                                    .getBootClasspathAsStrings(false)));
        }

//        FileCollection classpath = scope.getJavaClasspath(COMPILE_CLASSPATH, CLASSES);
        if (keepDefaultBootstrap) {
            classpath =
                    classpath.plus(
                            appVariantOutputContext.getVariantContext().getProject().files(appVariantOutputContext.getScope().getGlobalScope().getAndroidBuilder().getBootClasspath(false)));
        }
        javacTask.setClasspath(classpath);

        appVariantOutputContext.getScope().getTransformManager().addStream(OriginalStream.builder(appVariantOutputContext.getVariantContext().getProject(), "awb-classes")
                .addContentType(QualifiedContent.DefaultContentType.CLASSES)
                .addScope(QualifiedContent.Scope.PROJECT)
                .setFileCollection(appVariantOutputContext.getVariantContext().getProject().files(appVariantOutputContext.getJAwbavaOutputDir(awbBundle)))
                .build());
        appVariantOutputContext.getAwbTransformMap().get(awbBundle.getName()).addDir(appVariantOutputContext.getJAwbavaOutputDir(awbBundle));
        CompileOptions compileOptions = scope.getGlobalScope().getExtension().getCompileOptions();
        AbstractCompilesUtil.configureLanguageLevel(javacTask,
                compileOptions,
                scope.getGlobalScope()
                        .getExtension()
                        .getCompileSdkVersion(),
                scope.getJava8LangSupportType());
        javacTask.getOptions().setEncoding(compileOptions.getEncoding());

        Boolean includeCompileClasspath =
                scope.getVariantConfiguration()
                        .getJavaCompileOptions()
                        .getAnnotationProcessorOptions()
                        .getIncludeCompileClasspath();

        FileCollection processorPath =
                scope.getArtifactFileCollection(ANNOTATION_PROCESSOR, ALL, JAR);
        if (Boolean.TRUE.equals(includeCompileClasspath)) {
            // We need the jar files because annotation processors require the resources.
            processorPath = processorPath.plus(scope.getJavaClasspath(COMPILE_CLASSPATH, JAR));
        }

        javacTask.getOptions().setAnnotationProcessorPath(processorPath);


        boolean incremental = AbstractCompilesUtil.isIncremental(
                appVariantOutputContext.getScope().getGlobalScope().getProject(),
                scope,
                compileOptions,
                null, /* processorConfiguration, JavaCompile handles annotation processor now */
                LOG);

        javacTask.getOptions().setIncremental(incremental);


        AnnotationProcessorOptions annotationProcessorOptions =
                scope.getVariantConfiguration().getJavaCompileOptions()
                        .getAnnotationProcessorOptions();


//        javacTask.getOptions().getCompilerArgs().add("-processorpath");
//            javacTask.getOptions().getCompilerArgs().add(FileUtils.joinFilePaths(processorPath));

        if (!annotationProcessorOptions.getClassNames().isEmpty()) {
            javacTask.getOptions().getCompilerArgs().add("-processor");
            javacTask.getOptions().getCompilerArgs().add(
                    Joiner.on(',').join(annotationProcessorOptions.getClassNames()));
        }

        if (!annotationProcessorOptions.getArguments().isEmpty()) {
            for (Map.Entry<String, String> arg :
                    annotationProcessorOptions.getArguments().entrySet()) {

                String key = arg.getKey();
                String value = arg.getValue();

                if ("android.databinding.modulePackage".equals(key)) {
                    value = awbBundle.getPackageName() + "._bundleapp_";
                } else if ("android.databinding.artifactType".equals(key)) {
                    //value = "LIBRARY";
                } else if ("android.databinding.xmlOutDir".equals(key)) {
                    value = appVariantOutputContext.getVariantContext().getAwbLayoutInfoOutputForDataBinding(awbBundle)
                            .getAbsolutePath();
                } else if ("android.databinding.bindingBuildFolder".equals(key)) {
                    value = appVariantOutputContext.getVariantContext().getAwbDataBindingMergeArtifacts(awbBundle).getParentFile()
                            .getAbsolutePath();
                } else if ("android.databinding.generationalFileOutDir".equals(key)) {
                    value = value + "-" + awbBundle.getName();
                }

                javacTask.getOptions().getCompilerArgs().add(
                        "-A" + key + "=" + value);
            }
        }

        javacTask.getOptions().getCompilerArgs().add("-s");
        javacTask.getOptions().getCompilerArgs().add(
                appVariantOutputContext.getVariantContext().getAwbAnnotationProcessorOutputDir(awbBundle).getAbsolutePath());

        FileUtils.mkdirs(appVariantOutputContext.getVariantContext().getAwbAnnotationProcessorOutputDir(awbBundle));

        if (scope.getGlobalScope().getExtension().getDataBinding().isEnabled()
                && appVariantOutputContext.getVariantContext().isDataBindEnabled(awbBundle)) {
            File file = appVariantOutputContext.getVariantContext().getAwbDataBindingMergeArtifacts(awbBundle);
            ReflectUtils.updateField(javacTask, "dataBindingDependencyArtifacts", scope.getGlobalScope().getProject().files(file));



//            FileCollection files =
//                    appVariantOutputContext.getScope().getArtifactCollection(COMPILE_CLASSPATH, ALL, DATA_BINDING_ARTIFACT).getArtifactFiles();
//            for (File databingFile : files.getFiles()) {
//                for (File binFile : org.apache.commons.io.FileUtils.listFiles(databingFile,new String[]{"bin"},true)) {
//                    if (binFile.getName().contains("baseAdapters")) {
//                        try {
//                            FileUtils.copyFileToDirectory(binFile, appVariantOutputContext.getVariantContext().getAwbDataBindingMergeArtifacts(awbBundle));
//                        } catch (IOException e) {
//                            e.printStackTrace();
//                        }
//                    }
//                }
//            }
        }


            ReflectUtils.updateField(javacTask, "annotationProcessorOutputFolder", appVariantOutputContext.getVariantContext().getAwbAnnotationProcessorOutputDir(awbBundle));

            ReflectUtils.updateField(javacTask, "compileSdkVersion", scope.getGlobalScope().getExtension().getCompileSdkVersion());

            ReflectUtils.updateField(javacTask, "processorListFile", appVariantOutputContext.getVariantContext().getScope().getOutput(ANNOTATION_PROCESSOR_LIST));

            ReflectUtils.updateField(javacTask, "variantName", scope.getFullVariantName());

            ReflectUtils.updateField(javacTask, "mInstantRunBuildContext", scope.getInstantRunBuildContext());

            GradleBuildVariant.Builder builder = ProcessProfileWriter.getOrCreateVariant(scope.getGlobalScope().getProject().getPath(), scope.getFullVariantName());

            if (builder != null && builder.getAnnotationProcessorsList().size() > 0) {
                builder.clearAnnotationProcessors();
            }
            javacTask.setAwbBundle(awbBundle);

            //modification


    }

    private FileCollection getInputJars() {
        FileCollection classpath = scope.getJavaClasspath(AndroidArtifacts.ConsumedConfigType.COMPILE_CLASSPATH, AndroidArtifacts.ArtifactType.JAR);
        Set<File> dependencies = new HashSet<File>();
        dependencies.addAll(classpath.getFiles());
        //Increase the awb dependency

        dependencies.addAll(awbBundle.getLibraryJars());

        FileCollection allClassPatch = appVariantOutputContext.getVariantContext()
            .getProject()
            .files(dependencies);
        return allClassPatch;
    }
}
