/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.tasks.tpatch;

import com.alibaba.fastjson.JSON;
import com.android.build.gradle.api.BaseVariantOutput;
import com.android.build.gradle.internal.ApkDataUtils;
import com.android.build.gradle.internal.TaskContainerAdaptor;
import com.android.build.gradle.internal.api.ApContext;
import com.android.build.gradle.internal.api.AppVariantContext;
import com.android.build.gradle.internal.api.AppVariantOutputContext;
import com.android.build.gradle.internal.incremental.FileType;
import com.android.build.gradle.internal.incremental.InstantRunBuildContext;
import com.android.build.gradle.internal.scope.ConventionMappingHelper;
import com.android.build.gradle.internal.tasks.BaseTask;
import com.android.builder.signing.DefaultSigningConfig;
import com.android.builder.signing.SigningException;
import com.android.utils.Pair;
import com.intellij.util.io.ZipUtil;
import com.taobao.android.PatchManager;
import com.taobao.android.PatchType;
import com.taobao.android.builder.AtlasBuildContext;
import com.taobao.android.builder.dependency.model.AwbBundle;
import com.taobao.android.builder.extension.TBuildType;
import com.taobao.android.builder.insant.InstantInfo;
import com.taobao.android.builder.tasks.manager.MtlBaseTaskAction;
import com.taobao.android.builder.tools.BuildHelper;
import com.taobao.android.builder.tools.MD5Util;
import com.taobao.android.builder.tools.manifest.ManifestFileUtils;
import com.taobao.android.builder.tools.zip.ZipUtils;
import com.taobao.android.inputs.BaseInput;
import com.taobao.android.inputs.DexPatchInput;
import com.taobao.android.inputs.HotPatchInput;
import com.taobao.android.inputs.TpatchInput;
import com.taobao.android.object.ApkFileList;
import com.taobao.android.object.ArtifactBundleInfo;
import com.taobao.android.object.DiffType;
import com.taobao.android.tpatch.model.ApkBO;
import com.taobao.android.tpatch.model.BundleBO;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.dom4j.DocumentException;
import org.gradle.api.GradleException;
import org.gradle.api.tasks.Input;
import org.gradle.api.tasks.OutputDirectory;
import org.gradle.api.tasks.TaskAction;

import java.io.*;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.function.Consumer;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import static com.android.build.gradle.internal.api.ApContext.APK_FILE_MD5;

/**
 * Dynamic deployment of the tpatch build task
 * Created by wuzhong
 */
public class TPatchTask extends BaseTask {

    private TPatchContext patchContext;

    private DefaultSigningConfig signingConfig;

    private File outPatchFolder;

    private File instantInfoFile;

    private AppVariantContext appVariantContext;

    @TaskAction
    public void doTPatch() throws Exception {

        patchContext = getPatchContext();
        signingConfig = getSigningConfig();
        outPatchFolder = getOutPatchFolder();

        //the bundle List Copied to the outpatchFoulder
        new File(outPatchFolder, appVariantContext.bundleListCfg.getName()).delete();
        outPatchFolder.mkdirs();
        FileUtils.copyFileToDirectory(appVariantContext.bundleListCfg, outPatchFolder);

        // Get the container version
        String baseApkVersion = patchContext.getBaseVersionName();
        String newApkVersion = patchContext.versionName;

        File baseApk = patchContext.getBaseApk();

        File newApk = patchContext.diffApkFile;
        boolean retainMainBundleRes = true;

        if (null == newApk || !newApk.exists()) {
            newApk = patchContext.newApk;
            retainMainBundleRes = false;
        }

        getLogger().info("BaseApk:" +
                baseApk +
                ",baseVersion:" +
                baseApkVersion +
                ",newApk:" +
                newApk +
                ",newApkVersion:" +
                newApkVersion);

        ApkBO apkBO = new ApkBO(baseApk, baseApkVersion, baseApk.getName());
        ApkBO newApkBO = new ApkBO(newApk, newApkVersion, newApk.getName());
        BaseInput baseInput = createInput(apkBO, newApkBO, retainMainBundleRes);
        if (baseInput.patchType.equals(PatchType.DEXPATCH)) {
            List<ArtifactBundleInfo> modifyBundles = new ArrayList<ArtifactBundleInfo>();
            baseInput.artifactBundleInfos.forEach(artifactBundleInfo -> {
                if (artifactBundleInfo.getDiffType().equals(DiffType.MODIFY)) {
                    modifyBundles.add(artifactBundleInfo);
                }
            });
            if (modifyBundles.size() > 1) {
                throw new Exception("more than 1 bundle has changed:" + JSON.toJSONString(modifyBundles));
            }
        }
        PatchManager patchManager = new PatchManager(baseInput);
        patchManager.setLogger(getILogger());
        getLogger().info("start to do patch");

        patchManager.doPatch();

        getLogger().info("finish  do patch");


        try {

            FileUtils.writeStringToFile(new File(getOutPatchFolder(), "tpatch-bundles.json"),
                    JSON.toJSONString(patchContext.artifactBundleInfos));


            FileUtils.forceDelete(patchContext.newApk);

        } catch (Exception e) {
            throw new GradleException(e.getMessage(), e);
        }

    }


    private BaseInput createInput(ApkBO apkBO, ApkBO newApkBO, boolean retainMainBundleRes) throws IOException {
        TpatchInput tpatchInput = null;
        if (getProject().hasProperty("hotfix")) {
            tpatchInput = new HotPatchInput();
        } else {
            tpatchInput = new DexPatchInput();
        }
        tpatchInput.baseApkBo = apkBO;
        tpatchInput.newApkBo = newApkBO;
        tpatchInput.baseApkFileList = patchContext.getBaseApkFiles();
        tpatchInput.newApkFileList = patchContext.getNewApkFiles(appVariantContext);
        tpatchInput.outPatchDir = outPatchFolder;
        tpatchInput.productName = patchContext.appSignName;
        tpatchInput.outPutJson = new File(getOutPatchFolder(), "patchs.json");
        tpatchInput.artifactBundleInfos = patchContext.artifactBundleInfos;
        tpatchInput.diffBundleDex = true;
        tpatchInput.newPatch = patchContext.newPatch;
        tpatchInput.mainBundleName = patchContext.mainBundleName;
        tpatchInput.retainMainBundleRes = retainMainBundleRes;
        if (StringUtils.isNotBlank(patchContext.excludeFiles)) {
            tpatchInput.notIncludeFiles = (patchContext.excludeFiles.split(","));
        }
        if (apkBO.getVersionName().equals(newApkBO.getVersionName())) {
            if (tpatchInput instanceof HotPatchInput) {
                ((HotPatchInput) tpatchInput).hotClassListFile = patchContext.hotClassListFile;
                ((HotPatchInput) tpatchInput).patchType = PatchType.HOTFIX;
                ((HotPatchInput) tpatchInput).excludeClasses = patchContext.excludeClasses;

            } else {
                tpatchInput.patchType = PatchType.DEXPATCH;
                ((DexPatchInput) tpatchInput).patchClasses = patchContext.patchClasses;
                ((DexPatchInput) tpatchInput).excludeClasses = patchContext.excludeClasses;
            }
            tpatchInput.mainBundleName = "com.taobao.maindex";
        } else {
            tpatchInput.patchType = PatchType.TPATCH;
            tpatchInput.createHisPatch = true;
            tpatchInput.diffNativeSo = patchContext.diffNativeSo;
            tpatchInput.diffBundleSo = patchContext.diffBundleSo;
            tpatchInput.bundleWhiteList = appVariantContext.bundleListCfg;
            tpatchInput.createAll = StringUtils.isEmpty(patchContext.tpatchHistoryUrl);
            tpatchInput.LAST_PATCH_URL = patchContext.LAST_PATCH_URL;
            tpatchInput.hisPatchUrl = patchContext.tpatchHistoryUrl;
            if (null != patchContext.patchVersions) {
                tpatchInput.versionList = patchContext.patchVersions;
            }
        }
        List<Pair<BundleBO, BundleBO>> remoteBundles = new ArrayList<>();

        //Get the remote bundle
        for (AwbBundle awbBundle : AtlasBuildContext.awbBundleMap.values()) {
            if (awbBundle.isRemote) {

                File bundleFile = awbBundle.outputBundleFile;

                getProject().getLogger().error("add bundle compare " + bundleFile.getAbsolutePath());

                BundleBO newBundleBO = new BundleBO(awbBundle.getResolvedCoordinates().getArtifactId(), bundleFile, "");

                File baseBundleFile = new File(patchContext.apExplodeFolder, "remotebundles/" + bundleFile.getName());

                getProject().getLogger().error("add bundle compare base : " + baseBundleFile.getAbsolutePath());

                BundleBO baseBundleBO = null;
                if (baseBundleFile.exists()) {

                    getProject().getLogger().error(
                            "add bundle compare " + baseBundleFile.getAbsolutePath() + "->" + bundleFile.getAbsolutePath());

                    baseBundleBO = new BundleBO(awbBundle.getResolvedCoordinates().getArtifactId(), baseBundleFile, "");
                }
                remoteBundles.add(Pair.of(baseBundleBO, newBundleBO));
            }
        }

        if (remoteBundles.size() > 0) {
            tpatchInput.splitDiffBundle = remoteBundles;
        }
        return tpatchInput;

    }

    private void resignBaseApk(String baseApkVersion, ApkFileList apkFileList) throws IOException, SigningException {
        File baseVesrionApk = new File(patchContext.newApk.getParentFile(),
                patchContext.newApk.getName()
                        .replace(".apk", "-" + baseApkVersion + ".apk"));
        FileUtils.copyFile(patchContext.getBaseApk(), baseVesrionApk);
        if (patchContext.writeBuildInfo && StringUtils.isNotEmpty(patchContext.buildId)) {
            File buildFile = new File(getOutPatchFolder(), "build.txt");
            FileUtils.writeStringToFile(buildFile,
                    patchContext.buildId +
                            "," +
                            patchContext.versionName +
                            "," +
                            apkFileList.getMainBundle().get("classes.dex"));
            if (buildFile != null && buildFile.exists()) {
                getLogger().info("add build file to apk!");
                BuildHelper.writeFileToApk(buildFile, baseVesrionApk, "assets/build.txt");
            }

            String bundleInfoFileName = "bundleInfo-" +
                    appVariantContext.getVariantConfiguration().getVersionName() + ".json";
            File bundleInfoFile = new File(appVariantContext.getScope().getGlobalScope().getOutputsDir(),
                    bundleInfoFileName);
            if (bundleInfoFile.exists()) {
                getLogger().info("add " + bundleInfoFileName + " to apk!");
                BuildHelper.writeFileToApk(bundleInfoFile, baseVesrionApk, "assets/" + bundleInfoFileName);
            }

            BuildHelper.reSign(baseVesrionApk, signingConfig);
        }
    }

    public File getOutPatchFolder() {
        return outPatchFolder;
    }

    public DefaultSigningConfig getSigningConfig() {
        return signingConfig;
    }

    @Input
    public TPatchContext getPatchContext() {
        return patchContext;
    }

    public static class ConfigAction extends MtlBaseTaskAction<TPatchTask> {

        private AppVariantContext appVariantContext;

        public ConfigAction(AppVariantContext appVariantContext,
                            BaseVariantOutput baseVariantOutputData) {
            super(appVariantContext, baseVariantOutputData);
            this.appVariantContext = appVariantContext;
        }

        @Override
        public String getName() {
            return scope.getTaskName("createTPatch");
        }

        @Override
        public Class<TPatchTask> getType() {
            return TPatchTask.class;
        }

        @Override
        public void execute(TPatchTask tPatchTask) {

            super.execute(tPatchTask);

            final TBuildType tBuildType = appVariantContext.getBuildType();

            if (null == tBuildType ||
                    null == tBuildType.getPatchConfig() ||
                    !tBuildType.getPatchConfig().isCreateTPatch()) {
                tPatchTask.setEnabled(false);
                return;
            }


            tPatchTask.appVariantContext = appVariantContext;


            ConventionMappingHelper.map(tPatchTask, "outPatchFolder", new Callable<File>() {

                @Override
                public File call() throws Exception {
                    return getAppVariantOutputContext().getTPatchFolder();
                }
            });

            ConventionMappingHelper.map(tPatchTask,
                    "signingConfig",
                    new Callable<DefaultSigningConfig>() {

                        @Override
                        public DefaultSigningConfig call() throws Exception {
                            return appVariantContext.getSigningConfig();
                        }
                    });

            ConventionMappingHelper.map(tPatchTask, "patchContext", new Callable<TPatchContext>() {

                @Override
                public TPatchContext call() throws Exception {

                    TPatchContext tPatchContext = new TPatchContext();

                    AppVariantOutputContext appVariantOutputContext = getAppVariantOutputContext();

                    tPatchContext.diffApkFile = appVariantOutputContext.getDiffApk();
                    tPatchContext.newApk = appVariantOutputContext.getApkOutputFile(true);
                    tPatchContext.outPatchFolder = appVariantOutputContext.getTPatchFolder();
                    tPatchContext.manifestFile = new File(scope.getManifestProcessorTask().get(new TaskContainerAdaptor(scope.getGlobalScope().getProject().getTasks())).getManifestOutputDirectory(), "AndroidManifest.xml");
                    tPatchContext.apExplodeFolder = appVariantContext.apContext.getApExploredFolder();
                    tPatchContext.versionName = ApkDataUtils.get(baseVariantOutput).getVersionName();
                    tPatchContext.tpatchHistoryUrl = tBuildType.getPatchConfig()
                            .getTpatchHistoryUrl();
                    tPatchContext.hotClassListFile = tBuildType.getPatchConfig().getHotClassListFile();
                    tPatchContext.LAST_PATCH_URL = tBuildType.getPatchConfig().getLastPatchUrl();
                    tPatchContext.onlyBuildModifyAwb = tBuildType.getPatchConfig()
                            .getOnlyBuildModifyAwb();
                    tPatchContext.artifactBundleInfos = appVariantOutputContext.artifactBundleInfos;
                    tPatchContext.notPatchBundles = tBuildType.getPatchConfig().getNoPatchBundles();
                    tPatchContext.diffNativeSo = tBuildType.getPatchConfig().isDiffNativeSo();
                    tPatchContext.diffBundleSo = tBuildType.getPatchConfig().isDiffBundleSo();
                    tPatchContext.mainBundleName = tBuildType.getPatchConfig()
                            .getTpatchMainBundleName();
                    tPatchContext.excludeFiles = tBuildType.getPatchConfig()
                            .gettPatchNotIncludeFiles()
                            .isEmpty() ? "" : StringUtils.join(tBuildType.getPatchConfig()
                                    .gettPatchNotIncludeFiles(),
                            ",");

                    tPatchContext.buildId = tBuildType.getPatchConfig().getBuildId();

                    tPatchContext.newPatch = tBuildType.getPatchConfig().isNewPatch();
                    tPatchContext.writeBuildInfo = tBuildType.getPatchConfig()
                            .isTpatchWriteBuildInfo();
                    tPatchContext.diffBundleDex = tBuildType.getPatchConfig()
                            .isOnlyIncrementInAwb();
                    tPatchContext.diffMainDex = tBuildType.getPatchConfig()
                            .isOnlyIncrementInMain();
                    tPatchContext.excludeClasses = tBuildType.getPatchConfig()
                            .getExcludeClasses();

                    tPatchContext.patchClasses = tBuildType.getPatchConfig()
                            .getPatchClasses();

                    tPatchContext.appSignName = tBuildType.getPatchConfig().getAppSignName();

                    tPatchContext.patchVersions = tBuildType.getPatchConfig().getPatchVersions();

                    return tPatchContext;
                }
            });
        }
    }

    public static class TPatchContext implements Serializable {

        public File diffApkFile;

        public File newApk;

        public File outPatchFolder;

        public File manifestFile;

        public File apExplodeFolder;

        public String versionName;

        public String tpatchHistoryUrl;

        public String LAST_PATCH_URL;

        public Boolean onlyBuildModifyAwb;

        public String notPatchBundles;

        public String buildId;

        public boolean newPatch = true;

        public boolean diffNativeSo;

        public boolean writeBuildInfo;

        /**
         * Decide whether to diff the bundle's dex file
         */
        public boolean diffBundleDex;


        public Set<String> excludeClasses;

        public boolean diffMainDex;

        //    @Parameter(property = "android.patch.mainBundleName", defaultValue = "libcom_taobao_maindex")
        public String mainBundleName;

        public List<String> patchVersions;

        public Set<ArtifactBundleInfo> artifactBundleInfos;

        public Set<String> patchClasses = new HashSet<>();

        /**
         * patchDirectories that need to be excluded
         */
        public String excludeFiles;

        public String appSignName;

        public File hotClassListFile;

        public boolean diffBundleSo;

        public File getNewApkFiles(AppVariantContext appVariantContext) throws IOException {
            ApkFileList apkFileList = appVariantContext.getApkFiles().finalApkFileList;
            File apkFiles = new File(outPatchFolder.getParentFile(), APK_FILE_MD5);
            FileUtils.writeStringToFile(apkFiles, JSON.toJSONString(apkFileList));
            return apkFiles;
        }

        public File getBaseApk() {
            File apkFile = new File(apExplodeFolder, ApContext.AP_INLINE_APK_FILENAME);
            return apkFile;
        }

        public File getBaseApkFiles() {
            File apkFileList = new File(apExplodeFolder, APK_FILE_MD5);
            return apkFileList;
        }

        public String getBaseVersionName() throws IOException, DocumentException {
            File apManifestFile = new File(apExplodeFolder, "AndroidManifest.xml");
            String baseVersionName = ManifestFileUtils.getVersionName(apManifestFile);
            return baseVersionName;
        }
    }

    public static void zipPatch(File file, File dexFile) throws IOException {
        if (!file.exists())
            file.createNewFile();
        BufferedInputStream inputStream = new BufferedInputStream(new FileInputStream(dexFile));
        byte[] BUFFER = new byte[4096];
        FileOutputStream fOutputStream = new FileOutputStream(file);
        ZipOutputStream zoutput = new ZipOutputStream(fOutputStream);
        ZipEntry zEntry = new ZipEntry(dexFile.getName());
        zoutput.putNextEntry(zEntry);
        int len;
        while ((len = inputStream.read(BUFFER)) > 0) {
            zoutput.write(BUFFER, 0, len);
        }
        zoutput.closeEntry();
        zoutput.close();
        inputStream.close();
    }

}
