/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.tools.bundleinfo;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import android.aapt.pb.internal.ResourcesInternal.CompiledFile;
import com.android.annotations.NonNull;
import com.android.build.gradle.internal.api.ApkFiles;
import com.android.build.gradle.internal.api.AppVariantContext;
import com.google.common.base.Preconditions;
import com.google.protobuf.CodedInputStream;
import com.taobao.android.builder.AtlasBuildContext;
import com.taobao.android.builder.dependency.AtlasDependencyTree;
import com.taobao.android.builder.dependency.model.AwbBundle;
import com.taobao.android.builder.tools.MD5Util;
import com.taobao.android.builder.tools.PathUtil;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.IOFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.commons.lang.StringUtils;
import org.gradle.api.UncheckedIOException;
import org.gradle.api.tasks.StopExecutionException;

/**
 * Created by wuzhong on 16/6/15.
 */
public class ApkFileListUtils {

    /**
     * Records resource information for apk files
     */
    public static ApkFiles recordApkFileInfos(AppVariantContext appVariantContext) {

        ApkFiles apkFiles = new ApkFiles();

        List<File> mainBundleResFolders = new ArrayList<File>();
        mainBundleResFolders.add(appVariantContext.getScope().getVariantData().mergeResourcesTask.getOutputDir());
        prepareApkFileList(appVariantContext.getScope().getVariantData().mergeAssetsTask.getOutputDir(),
                           "assets", apkFiles);
        for (File resFolder : mainBundleResFolders) {
            prepareApkFileList(resFolder, "res", apkFiles);
        }

        // Record the resource information for each bundle
        AtlasDependencyTree dependencyTree = AtlasBuildContext.androidDependencyTrees.get(appVariantContext.getScope().
            getVariantConfiguration().getFullName());
        if (null == dependencyTree) {
            throw new StopExecutionException("DependencyTree cannot be null!");
        }

        List<AwbBundle> libs = dependencyTree.getAwbBundles();

        for (AwbBundle awbLib : libs) {
            File mergeAssertFolder = appVariantContext.getMergeAssets(awbLib);
            File mergeResFolder = appVariantContext.getMergeResources(awbLib);
            String awbName = awbLib.getName();
            prepareApkFileList(mergeAssertFolder, "assets", awbName, apkFiles);
            prepareApkFileList(mergeResFolder, "res", awbName, apkFiles);
        }

        return apkFiles;
    }

    /**
     * Initializes the file list information for the apk file
     *
     * @return
     */
    protected static void prepareApkFileList(File folder, String prefixName, ApkFiles apkFiles) {
        if (!folder.exists()) {
            return;
        }
        // Gets information about the main bundle
        Collection<File> files = FileUtils.listFiles(folder, new PureFileFilter(), TrueFileFilter.INSTANCE);
        for (File file : files) {
            if (file.isFile()) {
                String relativePath = prefixName + File.separator + PathUtil.toRelative(folder, file.getAbsolutePath());
                String md5;
                if ("res".equals(prefixName)) {
                    md5 = getResMd5(file);
                } else {
                    md5 = MD5Util.getFileMD5(file);
                }
                if (isImageFile(relativePath)) {
                    apkFiles.apkFileList.put(relativePath, md5);
                }
                apkFiles.finalApkFileList.put(relativePath, md5);
            }
        }
    }

    private static String getResMd5(File file) {
        String md5;
        if (isCompiledFile(file.getName())) {
            md5 = getCompiledFileMd5(file);
        } else {
            md5 = MD5Util.getFileMD5(file);
        }
        return md5;
    }

    private static String getCompiledFileMd5(File file) {
        String md5;
        try (InputStream is = new BufferedInputStream(new FileInputStream(file))) {
            CodedInputStream codedInputStream = CodedInputStream.newInstance(is);
            int numFiles = codedInputStream.readRawLittleEndian32();
            // Preconditions.checkState(numFiles == 1, "inline xml not implemented yet");
            long pbSize = codedInputStream.readRawLittleEndian64();
            codedInputStream.pushLimit((int)pbSize);
            CompiledFile compiledFile = CompiledFile.parser().parsePartialFrom(codedInputStream);
            md5 = MD5Util.getFileMD5(compiledFile.getSourcePath());
            // codedInputStream.popLimit((int)pbSize);
            // codedInputStream.pushLimit(0);
            return md5;
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    protected static void prepareApkFileList(File folder, String prefixName, String awbName, ApkFiles apkFiles) {
        if (!folder.exists()) {
            return;
        }
        // Gets information about the main bundle
        Collection<File> files = FileUtils.listFiles(folder, new PureFileFilter(), TrueFileFilter.INSTANCE);
        for (File file : files) {
            if (file.isFile()) {
                String relativePath = prefixName + File.separator + PathUtil.toRelative(folder, file.getAbsolutePath());
                String md5 = MD5Util.getFileMD5(file);
                if (isImageFile(relativePath)) {
                    if (null == apkFiles.apkFileList.getAwbs().get(awbName)) {
                        apkFiles.apkFileList.getAwbs().put(awbName, new HashMap<String, String>());
                    }
                    apkFiles.apkFileList.getAwbs().get(awbName).put(relativePath, md5);
                }
                if (null == apkFiles.finalApkFileList.getAwbs().get(awbName)) {
                    apkFiles.finalApkFileList.getAwbs().put(awbName, new HashMap<String, String>());
                }
                apkFiles.finalApkFileList.getAwbs().get(awbName).put(relativePath, md5);
            }
        }

    }

    private static final String[] IMG_EXTENSIONS = new String[] {".png", ".jpg"};

    private static boolean isImageFile(String name) {
        for (String imgExt : IMG_EXTENSIONS) {
            if (StringUtils.endsWithIgnoreCase(name, imgExt)) {
                return true;
            }
        }
        return false;
    }

    @NonNull
    private static final String FLAT_FILE_EXTENSION = ".flat";

    @NonNull
    private static final String ARSC_FLAT_FILE_EXTENSION = ".arsc.flat";

    private static boolean isCompiledFile(String name) {
        if (StringUtils.endsWithIgnoreCase(name, FLAT_FILE_EXTENSION) && !StringUtils.endsWithIgnoreCase(name,
            ARSC_FLAT_FILE_EXTENSION)) {
            return true;
        }
        return false;
    }

    public static class PureFileFilter implements IOFileFilter {
        @Override
        public boolean accept(File file) {
            return accept(file.getParentFile(), file.getName());
        }

        @Override
        public boolean accept(File dir, String name) {
            if (isIgnore(name)) {
                return false;
            }
            return !new File(dir, name).isDirectory();
        }
    }

    public static final String[] DEFAULTEXCLUDES = new String[] {"**/*~", "**/#*#", "**/.#*", "**/%*%", "**/._*",
        "**/CVS", "**/CVS/**", "**/.cvsignore", "**/RCS", "**/RCS/**", "**/SCCS", "**/SCCS/**", "**/vssver.scc",
        "**/.svn", "**/.svn/**", "**/.arch-ids", "**/.arch-ids/**", "**/.bzr", "**/.bzr/**", "**/.MySCMServerInfo",
        "**/.DS_Store", "**/.metadata", "**/.metadata/**", "**/.hg", "**/.hg/**", "**/.git", "**/.git/**",
        "**/BitKeeper", "**/BitKeeper/**", "**/ChangeSet", "**/ChangeSet/**", "**/_darcs", "**/_darcs/**",
        "**/.darcsrepo", "**/.darcsrepo/**", "**/-darcs-backup*", "**/.darcs-temp-mail"};

    private static boolean isIgnore(String name) {
        for (String pattern : DEFAULTEXCLUDES) {
            if (match(pattern, name, true)) {
                return true;
            }
        }
        return false;
    }

    public static boolean match(String pattern, String str, boolean isCaseSensitive) {
        char[] patArr = pattern.toCharArray();
        char[] strArr = str.toCharArray();
        int patIdxStart = 0;
        int patIdxEnd = patArr.length - 1;
        int strIdxStart = 0;
        int strIdxEnd = strArr.length - 1;
        boolean containsStar = false;

        int i;
        for (i = 0; i < patArr.length; ++i) {
            if (patArr[i] == 42) {
                containsStar = true;
                break;
            }
        }

        char ch;
        if (!containsStar) {
            if (patIdxEnd != strIdxEnd) {
                return false;
            } else {
                for (i = 0; i <= patIdxEnd; ++i) {
                    ch = patArr[i];
                    if (ch != 63 && !equals(ch, strArr[i], isCaseSensitive)) {
                        return false;
                    }
                }

                return true;
            }
        } else if (patIdxEnd == 0) {
            return true;
        } else {
            while ((ch = patArr[patIdxStart]) != 42 && strIdxStart <= strIdxEnd) {
                if (ch != 63 && !equals(ch, strArr[strIdxStart], isCaseSensitive)) {
                    return false;
                }

                ++patIdxStart;
                ++strIdxStart;
            }

            if (strIdxStart > strIdxEnd) {
                for (i = patIdxStart; i <= patIdxEnd; ++i) {
                    if (patArr[i] != 42) {
                        return false;
                    }
                }

                return true;
            } else {
                while ((ch = patArr[patIdxEnd]) != 42 && strIdxStart <= strIdxEnd) {
                    if (ch != 63 && !equals(ch, strArr[strIdxEnd], isCaseSensitive)) {
                        return false;
                    }

                    --patIdxEnd;
                    --strIdxEnd;
                }

                if (strIdxStart > strIdxEnd) {
                    for (i = patIdxStart; i <= patIdxEnd; ++i) {
                        if (patArr[i] != 42) {
                            return false;
                        }
                    }

                    return true;
                } else {
                    while (patIdxStart != patIdxEnd && strIdxStart <= strIdxEnd) {
                        i = -1;

                        int patLength;
                        for (patLength = patIdxStart + 1; patLength <= patIdxEnd; ++patLength) {
                            if (patArr[patLength] == 42) {
                                i = patLength;
                                break;
                            }
                        }

                        if (i == patIdxStart + 1) {
                            ++patIdxStart;
                        } else {
                            patLength = i - patIdxStart - 1;
                            int strLength = strIdxEnd - strIdxStart + 1;
                            int foundIdx = -1;
                            int i1 = 0;

                            label132:
                            while (i1 <= strLength - patLength) {
                                for (int j = 0; j < patLength; ++j) {
                                    ch = patArr[patIdxStart + j + 1];
                                    if (ch != 63 && !equals(ch, strArr[strIdxStart + i1 + j], isCaseSensitive)) {
                                        ++i1;
                                        continue label132;
                                    }
                                }

                                foundIdx = strIdxStart + i1;
                                break;
                            }

                            if (foundIdx == -1) {
                                return false;
                            }

                            patIdxStart = i;
                            strIdxStart = foundIdx + patLength;
                        }
                    }

                    for (i = patIdxStart; i <= patIdxEnd; ++i) {
                        if (patArr[i] != 42) {
                            return false;
                        }
                    }

                    return true;
                }
            }
        }
    }

    private static boolean equals(char c1, char c2, boolean isCaseSensitive) {
        return c1 == c2 ? true : !isCaseSensitive && (Character.toUpperCase(c1) == Character.toUpperCase(c2)
            || Character.toLowerCase(c1) == Character.toLowerCase(c2));
    }

}
