/*
 *
 *
 *
 *                                   Apache License
 *                             Version 2.0, January 2004
 *                          http://www.apache.org/licenses/
 *
 *     TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
 *
 *     1. Definitions.
 *
 *        "License" shall mean the terms and conditions for use, reproduction,
 *        and distribution as defined by Sections 1 through 9 of this document.
 *
 *        "Licensor" shall mean the copyright owner or entity authorized by
 *        the copyright owner that is granting the License.
 *
 *        "Legal Entity" shall mean the union of the acting entity and all
 *        other entities that control, are controlled by, or are under common
 *        control with that entity. For the purposes of this definition,
 *        "control" means (i) the power, direct or indirect, to cause the
 *        direction or management of such entity, whether by contract or
 *        otherwise, or (ii) ownership of fifty percent (50%) or more of the
 *        outstanding shares, or (iii) beneficial ownership of such entity.
 *
 *        "You" (or "Your") shall mean an individual or Legal Entity
 *        exercising permissions granted by this License.
 *
 *        "Source" form shall mean the preferred form for making modifications,
 *        including but not limited to software source code, documentation
 *        source, and configuration files.
 *
 *        "Object" form shall mean any form resulting from mechanical
 *        transformation or translation of a Source form, including but
 *        not limited to compiled object code, generated documentation,
 *        and conversions to other media types.
 *
 *        "Work" shall mean the work of authorship, whether in Source or
 *        Object form, made available under the License, as indicated by a
 *        copyright notice that is included in or attached to the work
 *        (an example is provided in the Appendix below).
 *
 *        "Derivative Works" shall mean any work, whether in Source or Object
 *        form, that is based on (or derived from) the Work and for which the
 *        editorial revisions, annotations, elaborations, or other modifications
 *        represent, as a whole, an original work of authorship. For the purposes
 *        of this License, Derivative Works shall not include works that remain
 *        separable from, or merely link (or bind by name) to the interfaces of,
 *        the Work and Derivative Works thereof.
 *
 *        "Contribution" shall mean any work of authorship, including
 *        the original version of the Work and any modifications or additions
 *        to that Work or Derivative Works thereof, that is intentionally
 *        submitted to Licensor for inclusion in the Work by the copyright owner
 *        or by an individual or Legal Entity authorized to submit on behalf of
 *        the copyright owner. For the purposes of this definition, "submitted"
 *        means any form of electronic, verbal, or written communication sent
 *        to the Licensor or its representatives, including but not limited to
 *        communication on electronic mailing lists, source code control systems,
 *        and issue tracking systems that are managed by, or on behalf of, the
 *        Licensor for the purpose of discussing and improving the Work, but
 *        excluding communication that is conspicuously marked or otherwise
 *        designated in writing by the copyright owner as "Not a Contribution."
 *
 *        "Contributor" shall mean Licensor and any individual or Legal Entity
 *        on behalf of whom a Contribution has been received by Licensor and
 *        subsequently incorporated within the Work.
 *
 *     2. Grant of Copyright License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        copyright license to reproduce, prepare Derivative Works of,
 *        publicly display, publicly perform, sublicense, and distribute the
 *        Work and such Derivative Works in Source or Object form.
 *
 *     3. Grant of Patent License. Subject to the terms and conditions of
 *        this License, each Contributor hereby grants to You a perpetual,
 *        worldwide, non-exclusive, no-charge, royalty-free, irrevocable
 *        (except as stated in this section) patent license to make, have made,
 *        use, offer to sell, sell, import, and otherwise transfer the Work,
 *        where such license applies only to those patent claims licensable
 *        by such Contributor that are necessarily infringed by their
 *        Contribution(s) alone or by combination of their Contribution(s)
 *        with the Work to which such Contribution(s) was submitted. If You
 *        institute patent litigation against any entity (including a
 *        cross-claim or counterclaim in a lawsuit) alleging that the Work
 *        or a Contribution incorporated within the Work constitutes direct
 *        or contributory patent infringement, then any patent licenses
 *        granted to You under this License for that Work shall terminate
 *        as of the date such litigation is filed.
 *
 *     4. Redistribution. You may reproduce and distribute copies of the
 *        Work or Derivative Works thereof in any medium, with or without
 *        modifications, and in Source or Object form, provided that You
 *        meet the following conditions:
 *
 *        (a) You must give any other recipients of the Work or
 *            Derivative Works a copy of this License; and
 *
 *        (b) You must cause any modified files to carry prominent notices
 *            stating that You changed the files; and
 *
 *        (c) You must retain, in the Source form of any Derivative Works
 *            that You distribute, all copyright, patent, trademark, and
 *            attribution notices from the Source form of the Work,
 *            excluding those notices that do not pertain to any part of
 *            the Derivative Works; and
 *
 *        (d) If the Work includes a "NOTICE" text file as part of its
 *            distribution, then any Derivative Works that You distribute must
 *            include a readable copy of the attribution notices contained
 *            within such NOTICE file, excluding those notices that do not
 *            pertain to any part of the Derivative Works, in at least one
 *            of the following places: within a NOTICE text file distributed
 *            as part of the Derivative Works; within the Source form or
 *            documentation, if provided along with the Derivative Works; or,
 *            within a display generated by the Derivative Works, if and
 *            wherever such third-party notices normally appear. The contents
 *            of the NOTICE file are for informational purposes only and
 *            do not modify the License. You may add Your own attribution
 *            notices within Derivative Works that You distribute, alongside
 *            or as an addendum to the NOTICE text from the Work, provided
 *            that such additional attribution notices cannot be construed
 *            as modifying the License.
 *
 *        You may add Your own copyright statement to Your modifications and
 *        may provide additional or different license terms and conditions
 *        for use, reproduction, or distribution of Your modifications, or
 *        for any such Derivative Works as a whole, provided Your use,
 *        reproduction, and distribution of the Work otherwise complies with
 *        the conditions stated in this License.
 *
 *     5. Submission of Contributions. Unless You explicitly state otherwise,
 *        any Contribution intentionally submitted for inclusion in the Work
 *        by You to the Licensor shall be under the terms and conditions of
 *        this License, without any additional terms or conditions.
 *        Notwithstanding the above, nothing herein shall supersede or modify
 *        the terms of any separate license agreement you may have executed
 *        with Licensor regarding such Contributions.
 *
 *     6. Trademarks. This License does not grant permission to use the trade
 *        names, trademarks, service marks, or product names of the Licensor,
 *        except as required for reasonable and customary use in describing the
 *        origin of the Work and reproducing the content of the NOTICE file.
 *
 *     7. Disclaimer of Warranty. Unless required by applicable law or
 *        agreed to in writing, Licensor provides the Work (and each
 *        Contributor provides its Contributions) on an "AS IS" BASIS,
 *        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 *        implied, including, without limitation, any warranties or conditions
 *        of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
 *        PARTICULAR PURPOSE. You are solely responsible for determining the
 *        appropriateness of using or redistributing the Work and assume any
 *        risks associated with Your exercise of permissions under this License.
 *
 *     8. Limitation of Liability. In no event and under no legal theory,
 *        whether in tort (including negligence), contract, or otherwise,
 *        unless required by applicable law (such as deliberate and grossly
 *        negligent acts) or agreed to in writing, shall any Contributor be
 *        liable to You for damages, including any direct, indirect, special,
 *        incidental, or consequential damages of any character arising as a
 *        result of this License or out of the use or inability to use the
 *        Work (including but not limited to damages for loss of goodwill,
 *        work stoppage, computer failure or malfunction, or any and all
 *        other commercial damages or losses), even if such Contributor
 *        has been advised of the possibility of such damages.
 *
 *     9. Accepting Warranty or Additional Liability. While redistributing
 *        the Work or Derivative Works thereof, You may choose to offer,
 *        and charge a fee for, acceptance of support, warranty, indemnity,
 *        or other liability obligations and/or rights consistent with this
 *        License. However, in accepting such obligations, You may act only
 *        on Your own behalf and on Your sole responsibility, not on behalf
 *        of any other Contributor, and only if You agree to indemnify,
 *        defend, and hold each Contributor harmless for any liability
 *        incurred by, or claims asserted against, such Contributor by reason
 *        of your accepting any such warranty or additional liability.
 *
 *     END OF TERMS AND CONDITIONS
 *
 *     APPENDIX: How to apply the Apache License to your work.
 *
 *        To apply the Apache License to your work, attach the following
 *        boilerplate notice, with the fields enclosed by brackets "[]"
 *        replaced with your own identifying information. (Don't include
 *        the brackets!)  The text should be enclosed in the appropriate
 *        comment syntax for the file format. We also recommend that a
 *        file or class name and description of purpose be included on the
 *        same "printed page" as the copyright notice for easier
 *        identification within third-party archives.
 *
 *     Copyright 2016 Alibaba Group
 *
 *     Licensed under the Apache License, Version 2.0 (the "License");
 *     you may not use this file except in compliance with the License.
 *     You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 *
 *
 */

package com.taobao.android.builder.tools.multidex.dex;

import com.android.dex.Dex;
import com.android.dex.DexIndexOverflowException;
import com.android.dx.command.dexer.DxContext;
import com.android.dx.merge.CollisionPolicy;
import com.taobao.android.builder.extension.MultiDexConfig;
import com.taobao.android.builder.tools.concurrent.ExecutorServicesHelper;
import com.taobao.android.builder.tools.multidex.FastMultiDexer;
import org.apache.commons.lang.StringUtils;
import org.gradle.api.GradleException;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Created by wuzhong on 2017/5/31.
 */
public class DexMerger {

    public static final String FASTMAINDEX_JAR = "fastmaindex";
    private static Logger logger = LoggerFactory.getLogger(DexMerger.class);

    private MultiDexConfig multiDexConfig;
    private List<File> fileList;

    private LinkedHashMap<File, Dex> jarDexMap = new LinkedHashMap<>();
    public List<Dex> dexList = new ArrayList<>();

    public DexMerger(MultiDexConfig multiDexConfig, Collection<File> fileList) throws IOException {

        this.multiDexConfig = multiDexConfig;
        this.fileList = fileList.stream().sorted(new FileComparator()).collect(Collectors.toList());

        for (File file : fileList) {
            Dex dex = new Dex(file);
            jarDexMap.put(file, dex);
        }

        dexList = new ArrayList<>(jarDexMap.values());

    }

    public DexMerger(MultiDexConfig multiDexConfig, Map<File, Dex> fileDexMap) {
        this.multiDexConfig = multiDexConfig;
        this.fileList = fileDexMap.keySet().stream().sorted(new FileComparator()).collect(Collectors.toList());

        for (File file : fileList) {
            Dex dex = fileDexMap.get(file);
            jarDexMap.put(file, dex);
        }

        dexList = new ArrayList<>(jarDexMap.values());

    }

    public List<DexGroup> group() {
        List<DexGroup> dexGroupList = new ArrayList<>();
        addDexByRule(dexGroupList);

        addDexLimited(dexGroupList, true);

        return dexGroupList;

    }

    public void executeMerge(File outDexFolder, List<DexGroup> dexGroupList) {

        ExecutorServicesHelper executorServicesHelper = new ExecutorServicesHelper("dexmerge", LoggerFactory
            .getLogger(FastMultiDexer.class), dexGroupList.size());
        List<Runnable> runnables = new ArrayList<>(dexGroupList.size());

        Dex[] mergedList = new Dex[dexGroupList.size()];
        for (int i = 0; i < dexGroupList.size(); i++) {
            final List<Dex> dexes = dexGroupList.get(i).dexs;
            final int index = i;
            runnables.add(new Runnable() {
                @Override
                public void run() {
                    try {
                        mergeDex(outDexFolder, dexes, index, mergedList);
                    } catch (Throwable e) {
                        throw new GradleException(e.getMessage(), e);
                    }
                }
            });
        }

        try {
            executorServicesHelper.execute(runnables);
        } catch (InterruptedException e) {
            throw new GradleException(e.getMessage(), e);
        }

        if (!dexList.isEmpty() ) {
            mergeSmallDexs(outDexFolder, mergedList);
        }

    }

    private void mergeSmallDexs(File outDexFolder, Dex[] mergedList) {
        List<DexGroup> dexGroupList = new ArrayList<>();
        for (int i = 0; i < mergedList.length; i++) {
            DexGroup dexGroup = new DexGroup();
            dexGroup.firstDex = (0 == i);
            dexGroup.addDex(mergedList[i]);
            dexGroupList.add(i, dexGroup);
        }

        addDexLimited(dexGroupList, false);

        executeMerge(outDexFolder, dexGroupList);
    }

    private void mergeDex(File outDexFolder, List<Dex> tmpList, int index, Dex[] mergedList) throws IOException {

        com.android.dx.merge.DexMerger dexMerger = new com.android.dx.merge.DexMerger(
            tmpList.toArray(new Dex[0]),
            CollisionPolicy.KEEP_FIRST,new DxContext());

        dexMerger.setCompactWasteThreshold(1024);

        Dex dex = dexMerger.merge();

        mergedList[index] = dex;

        if (dexList.isEmpty()) {
            logger.warn("dexCount of {}, methods {} , fields {}", index, dex.getTableOfContents().methodIds.size,
                        dex.getTableOfContents().fieldIds.size);

            String name = "classes" + (0 == index ? "" : String.valueOf(index + 1)) + ".dex";

            File dexFile = new File(outDexFolder, name);

            dex.writeTo(dexFile);
        }
    }

    private void addDexByRule(List<DexGroup> dexDtos) {

        DexGroup fistDto = new DexGroup();
        fistDto.firstDex = (true);
        dexDtos.add(0, fistDto);
        for (File file : fileList) {
            if (file.getParentFile().getName().startsWith(FASTMAINDEX_JAR)) {
                logger.warn(String.format("add %s to first dex!",file.getAbsolutePath()));
                Dex dex = jarDexMap.get(file);
                fistDto.addDex(dex);
                dexList.remove(dex);
            }
        }

        if (StringUtils.isEmpty(multiDexConfig.getDexSplitRules())) {
            return;
        }

        String[] rules = multiDexConfig.getDexSplitRules().split(";");
        for (int i = 0; i < rules.length; i++) {

            DexGroup dexDto = null;
            if (0 == i) {
                dexDto = fistDto;
            } else {
                dexDto = new DexGroup();
                dexDtos.add(i, dexDto);
            }

            String rule = rules[i];
            if (StringUtils.isEmpty(rule)) {
                continue;
            }
            String[] items = rule.split(",");
            for (File file : fileList) {
                if (match(items, file) || (0 == i && file.getAbsolutePath().contains(FASTMAINDEX_JAR))) {

                    logger.warn("FASTMULTIDEX put file " + file.getAbsolutePath() + " to dex " + i);

                    Dex dex = jarDexMap.get(file);
                    if (!dexDto.addDex(dex)) {
                        throw new DexIndexOverflowException(file.getAbsolutePath() + " can't add to dex" + i);
                    }
                    dexList.remove(dex);
                }
            }
        }

    }

    private void addDexLimited(List<DexGroup> dexDtos, boolean limited) {

        int dexCount = multiDexConfig.getDexCount();
        if (dexCount == 0) {
            dexCount = Integer.MAX_VALUE;
        }

        int index = 0;
        DexGroup dexDto = getDexDto(dexDtos, index);
        for (Dex dex : new ArrayList<>(dexList)) {
            //If you can't join in, you need to re-extend a dex
            while (!dexDto.addDex(dex)) {

                if (limited && index >= dexCount) {
                    return;
                }

                dexDto = getDexDto(dexDtos, index++);

            }
            dexList.remove(dex);
        }

    }

    @NotNull
    private DexGroup getDexDto(List<DexGroup> dexDtos, int index) {
        if (dexDtos.size() <= index) {
            DexGroup dexDto = new DexGroup();
            dexDtos.add(index, dexDto);
            return dexDto;
        }

        DexGroup dexDto = dexDtos.get(index);
        if (null == dexDto) {
            dexDto = new DexGroup();
            if (index == 0) {
                dexDto.firstDex = true;
            }
            dexDtos.add(index, dexDto);
        }
        return dexDto;
    }

    private boolean match(String[] items, File file) {
        for (String artiId : items) {
            if (file.getParentFile().getName().contains(artiId)) {
                return true;
            }
        }
        return false;
    }

}
